/* FILE: energy.cc                 -*-Mode: c++-*-
 *
 * Abstract energy class, derived from Oxs_Ext class.
 *
 */

#include "energy.h"

/* End includes */

#ifdef EXPORT_CALC_COUNT
void Oxs_Energy::FillCalcCountOutput(const Oxs_SimState& state)
{
  calc_count_output.cache.state_id = state.Id();
  calc_count_output.cache.value    = static_cast<REAL8m>(calc_count);
}
#endif // EXPORT_CALC_COUNT

// Constructors
Oxs_Energy::Oxs_Energy
( const char* name,     // Child instance id
  Oxs_Director* newdtr  // App director
  ) : Oxs_Ext(name,newdtr),calc_count(0)
{}

Oxs_Energy::Oxs_Energy
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr    // MIF block argument string
  ) : Oxs_Ext(name,newdtr,argstr),calc_count(0)
{}

//Destructor
Oxs_Energy::~Oxs_Energy() {}

// Default problem initializer routine.  This sets up default
// output objects, so any child that redefines this function
// will probably want to embed a call to this Init() inside
// the child specific version.
BOOL Oxs_Energy::Init()
{
  energy_sum_output.Setup(this,InstanceName(),"Energy","J",1,
			  &Oxs_Energy::UpdateStandardOutputs,
			  director);
  field_output.Setup(this,InstanceName(),"Field","A/m",1,
		     &Oxs_Energy::UpdateStandardOutputs,
		     director);
#ifdef EXPORT_CALC_COUNT
  calc_count_output.Setup(this,InstanceName(),"Calc count","",0,
			  &Oxs_Energy::FillCalcCountOutput,
			  director);
#endif // EXPORT_CALC_COUNT
  // Note: MS VC++ 6.0 requires fully qualified member names

  // Eventually, caching should be handled by controlling Tcl script.
  // Until then, request caching of scalar energy output by default.
  energy_sum_output.CacheRequestIncrement(1);

  calc_count=0;
  return 1;
}


// Standard output object update interface
void Oxs_Energy::UpdateStandardOutputs(const Oxs_SimState& state)
{
  if(state.Id()==0) { // Safety
    return;
  }
  Oxs_EnergyData oed(state);

  // Set aside dummy buffer space for oed to point to.
  // Array space inside these buffers will be allocated
  // as needed inside GetEnergyData(), and freed upon
  // exit from this function (when the buffers are destroyed
  // as part of automatic variable cleanup).  
  Oxs_MeshValue<REAL8m> dummy_energy;
  Oxs_MeshValue<ThreeVector> dummy_field;
  oed.energy_buffer = &dummy_energy;
  oed.field_buffer  = &dummy_field;

  GetEnergyData(state,oed);
}

///////////////////////////////////////////////////////////////////
// Energy and energy derivatives calculation function.  This
// should be "conceptually const."  The energy return is average
// energy density for the corresponding cell, in J/m^3.  The
// field is in A/m, pE_pt is in J/s.
//   There are two GetEnergy functions.  The first is a
// private, virtual member function of Oxs_Energy.  It takes as
// imports Oxs_MeshValue references into which to store the
// energy and H results.  The second is a non-virtual public
// member function, GetEnergyData, that passes in Oxs_MeshValue
// references that *may* be used to store the results, and
// separate parameters that return pointers to where the
// results actually were stored.  If output caching in enabled,
// the Oxs_Energy base class calls the virtual GetEnergy
// function with references to the output cache Oxs_MeshValue
// objects.  Otherwise the import buffer space is used.
//   NOTE: The pE_pt export is the partial derivative of energy
// with respect to time.  For most energy terms this will be
// 0.  It will only be non-zero if there is an explicit dependence
// on time, as for example with a time-varying applied field.
void Oxs_Energy::GetEnergyData
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 )
{
  if(oed.energy_buffer==NULL || oed.field_buffer==NULL) {
    // Bad input
    string msg=string("Oxs_EnergyData object in function"
		      " Oxs_Energy::GetEnergyData"
		      " contains NULL buffer pointers.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  if(state.Id()==0) {
    string msg=string("Programming error:"
		      " Invalid (unlocked) state detected"
		      " in Oxs_Energy::GetEnergyData.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  if(field_output.GetCacheRequestCount()>0) {
    field_output.cache.state_id=0;
    oed.field_buffer = &field_output.cache.value;
  }

  oed.energy_buffer->AdjustSize(state.mesh);
  oed.field_buffer->AdjustSize(state.mesh);

  ++calc_count;
  GetEnergy(state,oed);

  if(field_output.GetCacheRequestCount()>0) {
    field_output.cache.state_id=state.Id();
  }
  if(energy_sum_output.GetCacheRequestCount()>0) {
    if(oed.energy_sum.IsSet()) {
      energy_sum_output.cache.value=oed.energy_sum;
    } else {
      UINT4m size = state.mesh->Size();
      REAL8m energy_sum=0;
      const Oxs_MeshValue<REAL8m>& ebuf = *oed.energy;
      const Oxs_Mesh& mesh = *state.mesh;
      for(UINT4m i=0; i<size; ++i) {
	energy_sum += ebuf[i] * mesh.Volume(i);
      }
      energy_sum_output.cache.value=energy_sum;
      oed.energy_sum.Set(energy_sum); // Might as well set this field
      /// if we are doing the calculation anyway.
    }
    energy_sum_output.cache.state_id=state.Id();
  }
}
