/* FILE: mesh.h                 -*-Mode: c++-*-
 *
 * Abstract Oxs_Mesh class, derives from Oxs_Ext.
 *
 */

#ifndef _OXS_MESH
#define _OXS_MESH

#include "threevector.h"
#include "util.h"
#include "ext.h"

/* End includes */

template<class T> class Oxs_MeshValue;  // Forward declaration

////////////////////////////////////////////////////////////////////////
// Oxs_Mesh (generic mesh) class
class Oxs_Mesh : public Oxs_Ext {
protected:
  // Default file output using generic Oxs_Mesh interface
  void WriteOvfIrregular
  (Tcl_Channel channel,   // These have same meaning
   const char* title,     // as imports to WriteOvf().
   const char* desc,
   const char* valueunit,
   const char* datatype,
   const char* precision,
   const Oxs_MeshValue<ThreeVector>* vec,
   const Oxs_MeshValue<REAL8m>* scale=NULL,
   const ThreeVector* stephints=NULL // If non-null, then the values are
                                   /// written as stephints to the file.
) const;
  Oxs_Mesh(const char* name,Oxs_Director* newdtr);
  Oxs_Mesh(const char* name,Oxs_Director* newdtr,const char* argstr);
public:
  virtual ~Oxs_Mesh() {}

  virtual void GetBoundingBox(Oxs_Box& bbox) const =0;
  // Bounding region for mesh

  virtual UINT4m Size() const =0;  // Number of basis elements
  virtual REAL8m TotalVolume() const =0; // Sum of volumes of all basis elts
  virtual REAL8m Volume(UINT4m index) const=0; // Integral of basis elt 'index'
  virtual void Center(UINT4m index,ThreeVector& location) const =0;
  /// Center of mass of basis elt 'index'

  // Channel output.  Throws an exception on error.
  // Child classes may use WriteOvfIrregular() above to handle
  // irregular grid output.
  //   NOTE: Perhaps "valueunit" should be embedded into the
  // Oxs_MeshValue class???
  virtual void WriteOvf
  (Tcl_Channel channel,   // Output channel
   const char* title,     // Long filename or title
   const char* desc,      // Description to embed in output file
   const char* valueunit, // Field units, such as "A/m".
   const char* meshtype,  // Either "rectangular" or "irregular"
   const char* datatype,  // Either "binary" or "text"
   const char* precision, // For binary, "4" or "8";
                          ///  for text, a printf-style format
   const Oxs_MeshValue<ThreeVector>* vec,    // Vector array
   const Oxs_MeshValue<REAL8m>* scale=NULL   // Optional scaling for vec
                      /// Set scale to NULL to use vec values directly.
   ) const =0 ;

  // Version of WriteOvf that takes a filename instead of a Tcl_Channel.
  // This is for backwards compatibility and convenience.
  void WriteOvf
  (const char* filename,  // Name of output file
   const char* title,     // Long filename or title
   const char* desc,      // Description to embed in output file
   const char* valueunit, // Field units, such as "A/m".
   const char* meshtype,  // Either "rectangular" or "irregular"
   const char* datatype,  // Either "binary" or "text"
   const char* precision, // For binary, "4" or "8";
                          ///  for text, a printf-style format
   const Oxs_MeshValue<ThreeVector>* vec,    // Vector array
   const Oxs_MeshValue<REAL8m>* scale=NULL   // Optional scaling for vec
                      /// Set scale to NULL to use vec values directly.
   ) const;

  // Summing routines.  The terms are weighted by the basis element volumes.
  virtual REAL8m VolumeSum(const Oxs_MeshValue<REAL8m>& scalar) const;
  virtual ThreeVector VolumeSum(const Oxs_MeshValue<ThreeVector>& vec) const;
  virtual ThreeVector VolumeSum(const Oxs_MeshValue<ThreeVector>& vec,
				const Oxs_MeshValue<REAL8m>& scale) const;
  /// "scale" is elementwise scaling.

  // Extra precision summing routines.  These are the same as above,
  // except that intermediate sums are accumulated into Nb_Xpfloat
  // variables.
  virtual REAL8m VolumeSumXp(const Oxs_MeshValue<REAL8m>& scalar) const;
  virtual ThreeVector VolumeSumXp(const Oxs_MeshValue<ThreeVector>& vec) const;
  virtual ThreeVector VolumeSumXp(const Oxs_MeshValue<ThreeVector>& vec,
				  const Oxs_MeshValue<REAL8m>& scale) const;

  // Max angle routine.  Every mesh should have some concept of a
  // "neighborhood."  This routine should return the minimum dot
  // product between neighboring vectors across the mesh, for all
  // those vectors for which the corresponding entry in zero_check
  // is non-zero.
  //   If the vectors in the input Oxs_MeshValue<ThreeVector> array are
  // all unit vectors, then the max angle can be construed from
  // acos(mindot).
  //   NB: These routines are likely sensitive to boundary conditions.
  virtual REAL8m MinNeighborDot(const Oxs_MeshValue<ThreeVector>& vec,
				const Oxs_MeshValue<REAL8m>& zero_check)
    const =0;
};

#endif // _OXS_MESH




