/* FILE: meshvalue.h                 -*-Mode: c++-*-
 *
 * Oxs_MeshValue templated class, intended for use with the
 * Oxs_Mesh family.
 *
 * Note: To maintain a clean dependency hierarchy and prevent cyclic
 *   dependencies, this class does not maintain a lock on the mesh
 *   used to construct it.  Only the size of the mesh is retained,
 *   although a client may call Oxs_MeshValue<T>::CheckMesh() to
 *   check that the size is compatible with a given mesh.  If
 *   Oxs_MeshValue becomes more complex in the future, it may be
 *   necessary to add a pointer to the mesh that construction was
 *   based upon, in which case it will be the responsibility of the
 *   holder and/or user of the Oxs_MeshValue object to insure the
 *   correctness of that pointer.
 */

#ifndef _OXS_MESHVALUE
#define _OXS_MESHVALUE

#include "mesh.h"
#include "oxsexcept.h"

/* End includes */

template<class T> class Oxs_MeshValue {
 private:
  UINT4m size;
  T* arr;
  void Free();
 public:
  Oxs_MeshValue();
  Oxs_MeshValue(const Oxs_Mesh* mesh);
  ~Oxs_MeshValue();

  Oxs_MeshValue(const Oxs_MeshValue<T> &other);  // Copy constructor
  Oxs_MeshValue<T>& operator=(const Oxs_MeshValue<T> &other);
  Oxs_MeshValue<T>& operator+=(const Oxs_MeshValue<T> &other);
  Oxs_MeshValue<T>& operator-=(const Oxs_MeshValue<T> &other);
  Oxs_MeshValue<T>& operator*=(REAL8m mult);

  Oxs_MeshValue<T>& operator=(const T &value); // Fills out
  /// *this with size copies of value

  void Swap(Oxs_MeshValue<T>& other);

  void Release() { Free(); }  // Frees arr
  void AdjustSize(const Oxs_Mesh* newmesh);  // Compares current size
  /// to size of newmesh.  If different, deletes current arr and
  /// allocates new one compatible with newmesh.

  BOOL CheckMesh(const Oxs_Mesh* mesh) const;
  /// Returns true if current arr size is compatible with mesh.

  inline UINT4m Size() const { return size; }
  inline const T& operator[](UINT4m index) const;
  inline T& operator[](UINT4m index);
};

////////////////////////////////////////////////////////////////////////
// Include all definitions here.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

template<class T>
const T& Oxs_MeshValue<T>::operator[](UINT4m index) const
{
#ifndef NDEBUG
  if(index>=size) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
		"const T& Oxs_MeshValue<T>::operator[] const: "
		"Array out-of-bounds; index=%u, size=%u",
		index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
  return arr[index];
}

template<class T>
T& Oxs_MeshValue<T>::operator[](UINT4m index)
{
#ifndef NDEBUG
  if(index>=size) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
		"T& Oxs_MeshValue<T>::operator[]: "
		"Array out-of-bounds; index=%u, size=%u",
		index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
  return arr[index];
}

template<class T>
void Oxs_MeshValue<T>::Free()
{
  // Free value array
  if(arr!=NULL) { delete[] arr; arr=NULL; }
  size=0;
}

template<class T>
Oxs_MeshValue<T>::Oxs_MeshValue()
  : size(0), arr(NULL)
{}

template<class T>
Oxs_MeshValue<T>::Oxs_MeshValue(const Oxs_Mesh* mesh)
  : size(0), arr(NULL)
{
  if(mesh!=NULL) {
    // Allocate value array
    size=mesh->Size();
    arr = new T[size];
  }
}

template<class T>
Oxs_MeshValue<T>::~Oxs_MeshValue()
{
  Free();
}

template<class T>
Oxs_MeshValue<T>::Oxs_MeshValue(const Oxs_MeshValue<T> &other)
{ // Copy constructor.
  size=other.size;
  if(size>0) {
    arr = new T[size];
    for(UINT4m i=0;i<size;i++) arr[i]=other.arr[i];
  } else {
    arr = NULL;
  }
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator=(const Oxs_MeshValue<T>& other)
{
  if(this == &other) return *this;  // Nothing to do.

  if(size!=other.size) {
    Free();
    size=other.size;
    if(size>0) {
      arr = new T[size];
    } else {
      arr = NULL;  // Safety
    }
  }
  for(UINT4m i=0;i<size;i++) arr[i]=other.arr[i];
  return *this;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator+=(const Oxs_MeshValue<T>& other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
	     "Size mismatch in Oxs_MeshValue<T>::operator+=");
  }
  for(UINT4m i=0;i<size;i++) arr[i]+=other.arr[i];
  return *this;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator-=(const Oxs_MeshValue<T>& other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
	     "Size mismatch in Oxs_MeshValue<T>::operator-=");
  }
  for(UINT4m i=0;i<size;i++) arr[i]-=other.arr[i];
  return *this;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator*=(REAL8m mult)
{
  for(UINT4m i=0;i<size;i++) arr[i]*=mult;
  return *this;
}


template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator=(const T &value)
{
  for(UINT4m i=0;i<size;i++) arr[i]=value;
  return *this;
}

template<class T> BOOL
Oxs_MeshValue<T>::CheckMesh(const Oxs_Mesh* mesh) const
{ // Returns true if size is compatible with mesh
  if(mesh==NULL) return 0;
  return (mesh->Size() == size);
}

template<class T>
void Oxs_MeshValue<T>::Swap(Oxs_MeshValue<T>& other)
{ // Note: This works okay also if this==&other
  UINT4m tsize = size;  size = other.size;  other.size = tsize;
  T*     tarr  = arr;   arr  = other.arr;   other.arr  = tarr;
}

template<class T>
void Oxs_MeshValue<T>::AdjustSize(const Oxs_Mesh* newmesh)
{
  if(newmesh == NULL) {
    Free();
  } else {
    if(newmesh->Size() != size) {
      // Allocate new value array
      if(arr!=NULL) { delete[] arr; arr=NULL; }
      size=newmesh->Size();
      if(size>0) arr = new T[size];
    }
  }
}

#endif // _OXS_MESHVALUE
