/* FILE: ouput.cc                -*-Mode: c++-*-
 *
 * Output classes.  Each class providing output should embed an
 * instance of an output class for each output provided.
 *   Most of the real code is in the templated subclasses.  For
 * portability reasons, that code is in the header file, output.h.
 */

#include "director.h"
#include "output.h"

/* End includes */

Oxs_Output::Oxs_Output(int priority_)
  : director(NULL),
    cache_support(0), cache_request_count(0),
    priority(priority_)
{}

Oxs_Output::~Oxs_Output()
{ // Deregisters
  if(director!=NULL) director->DeregisterOutput(this);
}

BOOL Oxs_Output::Setup
(const char* owner_name_,
 const char* output_name_,
 const char* output_type_,
 const char* output_units_,
 BOOL cache_support_,
 Oxs_Director* director_)
{
  owner_name = owner_name_;
  output_name = output_name_;
  output_type = output_type_;
  output_units = output_units_;
  cache_support = cache_support_;

  cache_request_count = 0;  // Reset to 0.

  if(director_ != director) {
    if(director!=NULL) director->DeregisterOutput(this);
    director=director_;
    if(director!=NULL) director->RegisterOutput(this);
  }

#ifndef NDEBUG
  if(!director_->IsRegisteredOutput(this)) {
    OXS_THROW(Oxs_ProgramLogicError,
	      "Output object not registered.");
  }
#endif // NDEBUG

  return 1;
}

string Oxs_Output::LongName() const
{
  string longname = string(owner_name) + string(":")
    + string(output_name);
  return longname;
}

// Helper function for templated child constructors.  See the
// "Note" comment attached to the Oxs_ScalarOutput and
// Oxs_VectorFieldOutput constructors for an explanation of
// this initialization nonsense.
string Oxs_Output::GetDefaultOutputFormatString() const
{
  string label;
  string value;
  if(output_type.compare("scalar")==0) {
    label = "scalar_output_format";
  } else if(output_type.compare("vector field")==0) {
    label = "vector_field_output_format";
  } else {
    string msg = string("Output type \"") + output_type
      + string("\" is not recognized.");
    OXS_THROW(Oxs_ProgramLogicError,msg);
  }

  if(director==NULL) return string("");

  if(director->GetDriverAttributeValue(label,value)!=0) {
    string msg = string("Unable to obtain default output format"
			" of type \"") + label + string("\"");
    OXS_THROW(Oxs_ProgramLogicError,msg);
  }
  return value;
}

// Same "Note" as for Oxs_Output::GetDefaultOutputFormatString above.
const char* Oxs_Output::GetProblemName() const {
  return director->GetProblemName();
}

// Cache flag control
BOOL Oxs_Output::ForceCacheIncrement(INT4m incr)
{
  if(incr>0) cache_request_count+=incr;
  else if(incr<0) {
    UINT4m decr = -1*incr;
    if(cache_request_count < decr) cache_request_count = 0;
    else                           cache_request_count -= decr;
  }
  return 1;
}






