/* FILE: tclcommand.cc                 -*-Mode: c++-*-
 *
 * Wrapper for Tcl_Eval().
 *
 */

#include "tclcommand.h"
#include "util.h"

/* End includes */

Oxs_TclCommand::Oxs_TclCommand()
  : interp(NULL), no_too_few_restores_warning(0),
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
    eval_list_set(0),
#else
    base_command_size(0),objcmdptr(NULL),eval_result(NULL),
#endif // Tcl version check
    extra_args_count(0)
{}

Oxs_TclCommand::~Oxs_TclCommand()
{
  ReleaseBaseCommand();
}

void Oxs_TclCommand::ReleaseBaseCommand()
{
  extra_args_count=0;
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  eval_list_set=0;
  eval_list.Release();
  eval_result.erase();
  command_base.erase();
  command_args.clear();
#else
  if(eval_result!=NULL) {
    Tcl_DecrRefCount(eval_result);
    eval_result=NULL;
  }
  if(objcmdptr!=NULL) {
    delete objcmdptr;
    objcmdptr=NULL;
  }
  base_command_size=0;
#endif // Tcl version check

  // Empty result stack.  We discard all stacked results, on the
  // theory that if the destructor is being called with items still
  // in the stack, then we are probably in an exceptional (read,
  // error) state, and perhaps the current Tcl result holds some
  // useful error information.  The alternative is to dump all
  // the results except the one at the bottom of the stack,
  // with a command like
  //
  // if(it == result_stack.rend()-1) Tcl_RestoreResult(interp,&(*it));
  //
  // The boolean "unbalanced" is set true if the result stack is
  // not empty.  This presumably indicates a programming error.
  // After freeing resources, we print a warning if unbalanced
  // is true, provided an exception is not currently being processed,
  // and also provided an earlier exception did not explicitly
  // request no warning by setting the member variable
  // no_too_few_restores_warning true.
  vector<Tcl_SavedResult>::reverse_iterator it = result_stack.rbegin();
  BOOL unbalanced = (it != result_stack.rend());
  while(it != result_stack.rend()) {
    Tcl_DiscardResult(&(*it));
    ++it;
  }
  result_stack.clear();

  interp=NULL;

  if(unbalanced && !uncaught_exception()
     && !no_too_few_restores_warning) {
    PlainWarning("%s --- Unbalanced Tcl save/restore result calls"
		 " detected in Oxs_TclCommand: too few restores.",
		 exception_prefix.c_str());
  }

  no_too_few_restores_warning = 0; // Reset in any case

  exception_prefix.erase();
}

void Oxs_TclCommand::SetBaseCommand
(const char* exception_prefix_,
 Tcl_Interp* interp_,
 const string& cmdbase,
 int extra_args_count_)
{
  // Free old command, if any
  ReleaseBaseCommand();

  // Setup new command
  exception_prefix = exception_prefix_;
  interp = interp_;
  extra_args_count = extra_args_count_;
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  // Tcl string interface.
  command_base = cmdbase;
  for(int iarg=0;iarg<extra_args_count;iarg++) {
    command_args.push_back(string("")); // Dummy fill
  }
#else
  // Tcl obj interface
  Oxs_SplitList cmdlist;
  cmdlist.Split(cmdbase);
  base_command_size = cmdlist.Count();
  objcmdptr = new Oxs_TclObjArray(base_command_size+extra_args_count);
  for(int iarg=0;iarg<base_command_size;iarg++) {
    objcmdptr->WriteString(iarg,cmdlist[iarg]);
  }
#endif // Tcl version check
}

void Oxs_TclCommand::SaveInterpResult() const
{ // Conceptually const
  result_stack.resize(result_stack.size()+1);
  Tcl_SaveResult(interp,&(result_stack.back()));
}

void Oxs_TclCommand::RestoreInterpResult() const
{ // Conceptually const
  if(result_stack.size()<1) {
    string msg = exception_prefix
      + string(" --- Tcl result restore call"
	       " with no interp result saved.");
    OXS_THROW(Oxs_ProgramLogicError,msg);
  }
  Tcl_RestoreResult(interp,&(result_stack.back()));
  result_stack.pop_back();
}

void Oxs_TclCommand::DiscardInterpResult() const
{ // Conceptually const
  if(result_stack.size()<1) {
    string msg = exception_prefix
      + string(" --- Tcl result discard call"
	       " with no interp result saved.");
    OXS_THROW(Oxs_ProgramLogicError,msg);
  }
  Tcl_DiscardResult(&(result_stack.back()));
  result_stack.pop_back();
}

void Oxs_TclCommand::SetCommandArg(int index,const char* arg) const
{ // Conceptually const
#ifndef NDEBUG
  if(index>=extra_args_count || index<0) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::SetCommandArg: "
		"Array out-of-bounds; index=%d, arg=%.3000s",
		exception_prefix.c_str(),index,arg);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  command_args[index] = arg;
#else
  objcmdptr->WriteString(base_command_size+index,arg);
#endif // Tcl version check
}

void Oxs_TclCommand::SetCommandArg(int index,INT4m arg) const
{ // Conceptually const
#ifndef NDEBUG
  if(index>=extra_args_count || index<0) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::SetCommandArg: "
		"Array out-of-bounds; index=%d, arg=%ld",
		exception_prefix.c_str(),index,arg);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%ld",static_cast<long>(arg));
  command_args[index] = buf;
#else
  objcmdptr->WriteLong(base_command_size+index,static_cast<long>(arg));
#endif // Tcl version check
}

void Oxs_TclCommand::SetCommandArg(int index,UINT4m arg) const
{ // Conceptually const
#ifndef NDEBUG
  if(index>=extra_args_count || index<0) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::SetCommandArg: "
		"Array out-of-bounds; index=%d, arg=%ld",
		exception_prefix.c_str(),index,arg);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%lu",static_cast<unsigned long>(arg));
  command_args[index] = buf;
#else
  objcmdptr->WriteLong(base_command_size+index,static_cast<long>(arg));
  /// NOTE: there is no unsigned Tcl obj.  We could overflow
  /// check, but for now we'll just assume arg will fit into a long.
#endif // Tcl version check
}

void Oxs_TclCommand::SetCommandArg(int index,REAL8m arg) const
{ // Conceptually const
#ifndef NDEBUG
  if(index>=extra_args_count || index<0) {
    char buf[1024];
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::SetCommandArg: "
		"Array out-of-bounds; index=%d, arg=%.17g",
		exception_prefix.c_str(),index,arg);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%.17g",arg);
  command_args[index] = buf;
#else
  objcmdptr->WriteDouble(base_command_size+index,
			 static_cast<double>(arg));
#endif // Tcl version check
}


void Oxs_TclCommand::Eval() const
{ // Conceptually const
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  // For Tcl eval string interface
  if(eval_list_set) {
    eval_list.Release();
    eval_list_set=0;
  }
  string script = command_base
    + string(" ") + Oxs_MergeList(command_args);
  int errcode = Tcl_Eval(interp,Oc_AutoBuf(script.c_str()));
  eval_result = Tcl_GetStringResult(interp);
#else // ! (Tcl version < 8.1)
  // Use Tcl eval obj interface
  if(eval_result!=NULL) {
    Tcl_DecrRefCount(eval_result); // Free old result
    eval_result=NULL; // Safety
  }
  int errcode = Tcl_EvalObjv(interp,
			     objcmdptr->Size(),objcmdptr->Array(),
			     TCL_EVAL_GLOBAL);
  eval_result = Tcl_GetObjResult(interp);
  Tcl_IncrRefCount(eval_result);
#endif // (Tcl version < 8.1)
  if(errcode!=TCL_OK) {
    string msg = exception_prefix
      + string(" --- Error evaluating Tcl script: ");
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
    msg += script + string(" --- Error message: ") + eval_result;
#else
    msg += Oxs_MergeList(*objcmdptr)
      + string(" --- Error message: ")
      + string(Tcl_GetString(eval_result));
#endif
    // NOTE TO SELF: Change this to 0 to test Don's improved
    // DisplayError proc in oxsi.tcl (mjd, 25-June-2002)
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// This is expected to unbalance the results stack, unless the
    /// client wrapped up the Eval call in a try block.  There does
    /// not seem to be any real reason to print an unbalanced result
    /// stack warning too.
    OXS_THROW(Oxs_TclError,msg);
  }
}

string Oxs_TclCommand::GetWholeResult() const
{
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  return eval_result;
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  return string(Tcl_GetString(eval_result));
#endif // (Tcl version < 8.1)
}

#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
void Oxs_TclCommand::FillEvalList() const
{
  if(eval_list.Split(eval_result.c_str())!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + eval_result;
    OXS_THROW(Oxs_TclError,msg);
  }
  eval_list_set=1;
}
#endif

int Oxs_TclCommand::GetResultListSize() const
{
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  if(!eval_list_set) FillEvalList();
  return eval_list.Count();
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  int size;
  if(Tcl_ListObjLength(NULL,eval_result,&size)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + GetWholeResult();
    OXS_THROW(Oxs_TclError,msg);
  }
  return size;
#endif // (Tcl version < 8.1)
}

void Oxs_TclCommand::GetResultListItem(int index,string& result) const
{
#ifndef NDEBUG
  int size = GetResultListSize();
  if(index>=size || index<0) {
    char buf[1024];
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d"
		" (should be >0 and <%d).",
		exception_prefix.c_str(),index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  if(!eval_list_set) FillEvalList();
  result = eval_list[index];
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  Tcl_Obj* obj;
  if(Tcl_ListObjIndex(NULL,eval_result,index,&obj)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + GetWholeResult();
    OXS_THROW(Oxs_TclError,msg);
  }
  if(obj==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d.",
		index);
    OXS_THROW(Oxs_BadIndex,buf);
  }
  result = Tcl_GetString(obj);
#endif // (Tcl version < 8.1)
}

void Oxs_TclCommand::GetResultListItem(int index,INT4m& result) const
{
#ifndef NDEBUG
  int size = GetResultListSize();
  if(index>=size || index<0) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d"
		" (should be >0 and <%d).",
		exception_prefix.c_str(),index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  if(!eval_list_set) FillEvalList();
  char* cptr;
  result = static_cast<INT4m>(strtol(eval_list[index],&cptr,10));
  if(eval_list[index][0]=='\0' || *cptr!='\0') {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[2048];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Tcl script return item at index=%d"
		" is not an integer: %.500s",
		exception_prefix.c_str(),index,eval_list[index]);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  Tcl_Obj* obj;
  if(Tcl_ListObjIndex(NULL,eval_result,index,&obj)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + GetWholeResult();
    OXS_THROW(Oxs_TclError,msg);
  }
  if(obj==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d.",
		exception_prefix.c_str(),index);
    OXS_THROW(Oxs_BadIndex,buf);
  }
  long lresult;
  if(Tcl_GetLongFromObj(NULL,obj,&lresult)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[2048];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Tcl script return item at index=%d"
		" is not a long integer: %.500s",
		exception_prefix.c_str(),index,Tcl_GetString(obj));
    OXS_THROW(Oxs_BadIndex,buf);
  }
  result = static_cast<INT4m>(lresult);
#endif // (Tcl version < 8.1)
}

void Oxs_TclCommand::GetResultListItem(int index,UINT4m& result) const
{
#ifndef NDEBUG
  int size = GetResultListSize();
  if(index>=size || index<0) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d"
		" (should be >0 and <%d).",
		exception_prefix.c_str(),index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  if(!eval_list_set) FillEvalList();
  char* cptr;
  result = static_cast<UINT4m>(strtoul(eval_list[index],&cptr,10));
  if(eval_list[index][0]=='\0' || *cptr!='\0') {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[2048];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Tcl script return item at index=%d"
		" is not an integer: %.500s",
		exception_prefix.c_str(),index,eval_list[index]);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  Tcl_Obj* obj;
  if(Tcl_ListObjIndex(NULL,eval_result,index,&obj)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + GetWholeResult();
    OXS_THROW(Oxs_TclError,msg);
  }
  if(obj==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d.",
		exception_prefix.c_str(),index);
    OXS_THROW(Oxs_BadIndex,buf);
  }
  long lresult;
  if(Tcl_GetLongFromObj(NULL,obj,&lresult)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[2048];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Tcl script return item at index=%d"
		" is not a long integer: %.500s",
		exception_prefix.c_str(),index,Tcl_GetString(obj));
    OXS_THROW(Oxs_BadIndex,buf);
  }
  result = static_cast<UINT4m>(lresult);
  // Note: There is no unsigned Tcl obj.
#endif // (Tcl version < 8.1)
}

void Oxs_TclCommand::GetResultListItem(int index,REAL8m& result) const
{
#ifndef NDEBUG
  int size = GetResultListSize();
  if(index>=size || index<0) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d"
		" (should be >0 and <%d).",
		exception_prefix.c_str(),index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  if(!eval_list_set) FillEvalList();
  BOOL err;
  result = static_cast<REAL8m>(Nb_Atof(eval_list[index],err));
  if(err) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[2048];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Tcl script return item at index=%d"
		" is not a floating point number: %.500s",
		exception_prefix.c_str(),index,eval_list[index]);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  Tcl_Obj* obj;
  if(Tcl_ListObjIndex(NULL,eval_result,index,&obj)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + GetWholeResult();
    OXS_THROW(Oxs_TclError,msg);
  }
  if(obj==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Array out-of-bounds; index=%d.",
		exception_prefix.c_str(),index);
    OXS_THROW(Oxs_BadIndex,buf);
  }
  double tmpresult;
  if(Tcl_GetDoubleFromObj(NULL,obj,&tmpresult)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    char buf[2048];
    Oc_Snprintf(buf,sizeof(buf),
		"%.800s --- void Oxs_TclCommand::GetResultListItem:"
		" Tcl script return item at index=%d"
		" is not a floating point number: %.500s",
		exception_prefix.c_str(),index,Tcl_GetString(obj));
    OXS_THROW(Oxs_BadIndex,buf);
  }
  result = static_cast<REAL8m>(tmpresult);
#endif // (Tcl version < 8.1)
}

void Oxs_TclCommand::GetResultList(vector<string>& result_list) const
{
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  if(!eval_list_set) FillEvalList();
  eval_list.FillParams(result_list);
#else // ! (Tcl version < 8.1)
  if(eval_result==NULL) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    OXS_THROW(Oxs_ProgramLogicError,"No eval result;"
	      " Eval() has not been called on Oxs_TclCommand");
  }
  int obj_count;
  Tcl_Obj** obj_arr;
  if(Tcl_ListObjGetElements(NULL,eval_result,&obj_count,&obj_arr)!=TCL_OK) {
    no_too_few_restores_warning = 1; // We're throwing an exception.
    /// See notes in ::Eval() about result stack and exceptions
    string msg = exception_prefix
      + string(" --- Tcl script return is not a list: ")
      + GetWholeResult();
    OXS_THROW(Oxs_TclError,msg);
  }
  result_list.clear();
  for(int i=0;i<obj_count;i++) {
    result_list.push_back(string(Tcl_GetString(obj_arr[i])));
  }
#endif // (Tcl version < 8.1)
}

/////////////////////////////////////////////////////////////////////
//
// Oxs_ParseTclCommandLineRequest, intended to be used in support of
// Oxs_TclCommand. The label and param_count fields of each
// Oxs_TclCommandLineOption object in the import/export "options"
// parameter should be filled on entry. Import request_string is
// a Tcl list, where each list item matches the label portion of
// some Oxs_TclCommandLineOption object in "options".
// Oxs_ParseTclCommandLineRequest fills in the position field of
// each Oxs_TclCommandLineOption object corresponding to the
// position of that option in the request_string.  If the option
// is not used, the position field is set to -1.  The position
// is padded to account for the number of parameters in each
// preceding option in request_string.  The return value is the
// number of total parameters associated with the options selected
// in request_string.  Exceptions are thrown if request_string
// contains an option not included in "options", or if the same
// options is selected more than once.
//
int Oxs_ParseTclCommandLineRequest
(const char* exception_prefix,
 vector<Oxs_TclCommandLineOption>& options,
 const string& request_string)
{
  // Initialize options vector
  vector<Oxs_TclCommandLineOption>::iterator opit = options.begin();
  while(opit != options.end()) {
    opit->position = -1;
    ++opit;
  }

  // Split request string
  Oxs_SplitList request;
  request.Split(request_string);

  // Process request string
  int total_param_count=0;
  for(int i=0;i<request.Count();i++) {
    opit = options.begin();
    while(opit != options.end()) {
      if(opit->Label().compare(request[i])==0) {
	// Match
	if(opit->position >= 0) {
	  // Option already selected
	  string msg = string(exception_prefix)
	    + string(" --- Duplicate option request: \"")
	    + request[i] + string("\"");
	  OXS_THROW(Oxs_BadParameter,msg);
	}
	opit->position = total_param_count;
	total_param_count += opit->ParamCount();
	break;
      }
      ++opit;
    }
    if(opit == options.end()) {
      // No match
      string msg = string(exception_prefix)
	+ string("Unrecognized option request: \"")
	+ request[i] + string("\"");
      OXS_THROW(Oxs_BadParameter,msg);
    }
  }

  return total_param_count;
}
