/* FILE: util.h                 -*-Mode: c++-*-
 *
 * Small utility classes
 *
 */

#ifndef _OXS_UTIL
#define _OXS_UTIL

#include <stdio.h>
#include <string>
#include <vector>
#include "oc.h"
#include "nb.h" // This is needed in util.cc.  We include it here
/// to get it into the dependency tree for overaggressive compilers
/// that when loading util.h automatically go looking into util.cc
/// too, presumably for template expansion.  Which compilers?  I'm
/// not naming names, but their initials are s, g and i.

#include "threevector.h"
#include "oxsexcept.h"

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported

/* End includes */

////////////////////////////////////////////////////////////////////////
// Wrapper for Tcl_SplitList().  This class makes sure memory allocated
// by Tcl_SplitList() is released, even if normal execution path is
// interrupted by an exception.
class Oxs_SplitList
{
private:
  int argc;
  CONST84 char ** argv;
public:
  Oxs_SplitList() : argc(0), argv(NULL) {}
  void Release();
  int Split(const char* argstr);
  int Split(const string& argstr) { return Split(argstr.c_str()); }
  int Count() const { return argc; }
  const char* operator[](int n) const;
  void FillParams(vector<string>& params) const;
  ~Oxs_SplitList();
};

#if (TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0))
////////////////////////////////////////////////////////////////////////
// Wrapper for Tcl_Obj**
typedef Tcl_Obj* OxsTclObjPtr;
typedef OxsTclObjPtr& OxsTclObjPtrRef;
typedef const Tcl_Obj* OxsTclObjConstPtr;
typedef OxsTclObjConstPtr const & OxsTclObjConstPtrConstRef;
class Oxs_TclObjArray
{
private:
  int size;
  Tcl_Obj** arr;
public:
  Oxs_TclObjArray(int arrsize);
  ~Oxs_TclObjArray();

  OxsTclObjPtrRef operator[](int index);
  OxsTclObjConstPtrConstRef operator[](int index) const;
  /// It is not entirely clear to me whether the latter should return a
  /// constant reference to a constant pointer, or a constant reference
  /// to a non-constant pointer.  We'll try the more restrictive
  /// approach first. -mjd, 19-Feb-2002.

  int Size() const { return size; }
  Tcl_Obj** Array() { return arr; }

  // Write interface routines.  These handle the 
  // copy-on-write semantics of the Tcl referencing
  // system.
  void WriteString(int index,const char*);
  void WriteString(int index,const string& str);
  void WriteDouble(int index,double val);
  void WriteInt(int index,int val);
  void WriteLong(int index,long val);
  /// Add more as needed

  // Read interface routines.
  string GetString(int index) const;

};
#endif // Tcl version check

////////////////////////////////////////////////////////////////////////
// Wrappers for Tcl_Merge().
string Oxs_MergeList(const vector<string>* args);
string Oxs_MergeList(const vector<string>& args);
#if (TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0))
string Oxs_MergeList(const Oxs_TclObjArray& arr);
#endif // Tcl version check

////////////////////////////////////////////////////////////////////////
// Wrapper for standard C stream file pointer.  Constructor opens
// file, destructor closes it.
struct Oxs_FilePointer
{
public:
  Oxs_FilePointer(const char* filename,const char* filemode);
  ~Oxs_FilePointer();
  FILE* fptr;
};

////////////////////////////////////////////////////////////////////////
// Simple box class.  Used for defining axes parallel rectangular
// parallelepipeds.
class Oxs_Box {
  REAL8m x0,x1,y0,y1,z0,z1;  // Box corner coordinates.
public:
  Oxs_Box() : x0(0.),x1(-1.),y0(0.),y1(-1.),z0(0.),z1(-1.) {}

  void Set(REAL8m x0p,REAL8m x1p,REAL8m y0p,REAL8m y1p,
	   REAL8m z0p,REAL8m z1p);
  /// This Set function insures that x0<=x1, y0<=y1, z0<=z1.

  void Set(const string& sx0,const string& sx1,
	   const string& sy0,const string& sy1,
	   const string& sz0,const string& sz1);
  /// This Set function insures that x0<=x1, y0<=y1, z0<=z1.

  BOOL CheckOrderSet(REAL8m xmin,REAL8m xmax,
		     REAL8m ymin,REAL8m ymax,
		     REAL8m zmin,REAL8m zmax);
  /// Checks that [xyz]min <= [xyz]max.  Returns 0 if not,
  /// otherwise sets [xyz][01] and returns 1.

  // An empty box is identified by the condition x0>x1.
  void MakeEmpty() {
    x0 = y0 = z0 =  0.0;
    x1 = y1 = z1 = -1.0;
  }

  BOOL IsEmpty() const { return (x0>x1); }

  BOOL Expand(const Oxs_Box& other); // Expands *this as
  /// necessary to contain region specified by "other".
  /// Returns 1 if *this was enlarged, 0 if other was
  /// already contained in *this.

  BOOL Expand(REAL8m x,REAL8m y,REAL8m z); // Similar
  /// to Expand(const Oxs_Box&), but with a single
  /// point import instead of a box.  Returns 1 if *this
  /// was enlarged, 0 if (x,y,z) was already contained
  /// in *this.

  BOOL Intersect(const Oxs_Box& other); // Shrinks *this
  /// as necessary so that the resulting box is the
  /// intersection of the original *this with other.

  BOOL IsIn(const ThreeVector& point) const;
  BOOL IsContained(const Oxs_Box& other) const;

  REAL8m GetMinX() const { return x0; }
  REAL8m GetMaxX() const { return x1; }
  REAL8m GetMinY() const { return y0; }
  REAL8m GetMaxY() const { return y1; }
  REAL8m GetMinZ() const { return z0; }
  REAL8m GetMaxZ() const { return z1; }

  REAL8m GetVolume() const {
    if(IsEmpty()) return 0.0;
    return (x1-x0)*(y1-y0)*(z1-z0);
  }
};
// The next 3 operators are defined so MSVC++ 5.0 will accept
// vector<Oxs_Box>, but are left undefined because they aren't
// needed, and in general make no sense.
BOOL operator<(const Oxs_Box&,const Oxs_Box&);
BOOL operator>(const Oxs_Box&,const Oxs_Box&);
BOOL operator==(const Oxs_Box&,const Oxs_Box&);

////////////////////////////////////////////////////////////////////////
// Write Once Object (WOO) template.  Can be used with any type that
// supports operator=().  Throws an exception if a read occurs on an
// unset object, or if a write occurs onto an already set object.
// These type is not intended for use in high traffic paths.  If such
// a need arises, then perhaps the access checks should be #ifdef'd out
// (perhaps with NDEBUG?) for production builds.
//  UPDATE 5-Apr-2002: New class idea ---
//    Problems: I want to be able to store arbitrary types, with fast
//  lookup and memory reuse.
//    Answer: Use a 3-class solution.  First, a WOO class holding
//  an auto_ptr to a generic wrapper class.  Children of the wrapper
//  class are templated.  The director then holds an STL vector of WOO's.
//  Clients register with the director each WOO they want stored in
//  the state.  When a state is created, the WOO vector will be filled
//  out with the indicated objects, with NULL auto_ptr's.  So every state
//  will hold slots for each WOO, and in the same order.  The return from
//  the registration call is an index to the corresponding position in
//  the WOO vector.
//     Each WOO also holds the "name" of the object being held, which is
//  the same across all states.  Clients can do a lookup on the name,
//  and get back the index for future fast access.
//     A const WOO object which is not "set" will allow clients to grab
//  its auto_ptr, and will accept a new auto_ptr.  But once a new auto_ptr
//  is accepted, any future attempt to grab that auto_ptr or write a new
//  one will result in an exception.  The non-const interface allows this
//  without exception (or exceptions :^).
//     When a state is reused, all the WOO objects are marked as unset.
//  Memory is not released, though, so by grabbing the auto_ptr the
//  associated memory may be reused.  Hmm, the wrapper class should
//  have a virtual "reset" method that re-initializes the object.
//     Note 1: "grabbing" and auto_ptr, i.e., getting a non-const ptr,
//  is different from "reading" the auto_ptr, i.e., getting a const ptr.
//  The former raises an error if attempted on a set, const WOO object.
//  The latter raises an error if attempted on an unset object.  Also,
// "grabbing" unsets the WOO object.
//     Note 2: I've had problems using auto_ptr in the past due to
//  busted STL auto_ptr implementations.  Since auto_ptr is a simple
//  enough class, perhaps we should just go ahead and implement our
//  own stand-alone Oc_AutoPtr class?
//     Note 3: Maybe all state-based output should be moved into the
//  state?  Perhaps at some level of the 3-class WOO should include
//  all the machinery currently in the Oxs_Output objects, such as
//  update function pointer and output registration?  It might also
//  be useful to store output scheduling information here too, so that
//  when an output update function is called all other outputs that
//  will be requested for that state are calculated and cached(?) too?
//  Or maybe rather than caching, we store where the destinations of the
//  data, and send it straight out?
template<class T> class Oxs_WriteOnceObject
{
private:
  BOOL set;
  T value;
public:
  Oxs_WriteOnceObject() : set(0) {}
  BOOL IsSet() const { return set; }
  void Set(const T& newvalue) {
    if(set) {
      OXS_THROW(Oxs_ProgramLogicError,
		"WriteOnceObject<T>::Set(): Object already set");
    }
    value = newvalue;
    set = 1;
  }
  const T& Get() const {
    if(!set) {
      OXS_THROW(Oxs_ProgramLogicError,
		"WriteOnceObject<T>::Get(): Object not set");
    }
    return value;
  }
  const Oxs_WriteOnceObject<T>& operator=(const T& newvalue) {
    Set(newvalue);
    return *this;
  }
  operator const T&() const { return Get(); }
};


////////////////////////////////////////////////////////////////////////
// Owned pointer class.  A smart pointer with mediocre intelligence.
template<class T> class Oxs_OwnedPointer
{
private:
  BOOL owner;
  T* ptr;

  // Disable all copy operators by declaring them but not providing
  // a definition.  This makes OwnedPointers more awkward to use,
  // but less likely to be abused.
  Oxs_OwnedPointer(const Oxs_OwnedPointer<T>&);
  Oxs_OwnedPointer<T>& operator=(const Oxs_OwnedPointer<T>&);
public:
  Oxs_OwnedPointer() : owner(0), ptr(0) {}
  ~Oxs_OwnedPointer() { if(owner) delete ptr; }

  // Initialization
  void SetAsOwner(T* nptr) {
    if(owner) delete ptr;
    if((ptr=nptr)!=0) owner=1;
    else              owner=0; // Don't own NULL pointer
  }
  void SetAsNonOwner(T* nptr) {
    if(owner) delete ptr;
    owner=0;
    ptr=nptr;
  }
  void Free() { SetAsNonOwner(NULL); }

  // Dereferencing ops
  T* GetPtr() const { return ptr; }
  T* operator->() const { return ptr; }
  T& operator*() const { return *ptr; }

  // Ownership managment
  BOOL IsOwner() const { return owner; }

  void ReleaseOwnership() { owner=0; }
  /// Note difference with Free(): ReleaseOwnership()
  /// does *not* delete to pointed to object, or even
  /// stop pointing to it.

  void ClaimOwnership() {
    if(ptr==NULL) {
      OXS_THROW(Oxs_BadPointer,
		"Oxs_OwnedPointer<T>::ClaimOwnership():"
		" Can't take ownership of a NULL pointer.");
    }
    owner=1;
  }

};

////////////////////////////////////////////////////////////////////////
// Wrapper template for standard C++ new[]/delete[] memory allocation.
// Bare new's inside class constructors don't provide automatic cleanup
// (i.e., delete[] calls) if an exception is thrown inside the
// constructor.  On the other hand, STL container classes place copy
// constructor and operator= requirements on their templated class.
// This class provides a workaround for both problems.
template<class T> class Oxs_ArrayWrapper
{
private:
  UINT4m size;
  T* objs;

  // Disable copy operations by declaring the following
  // without providing definitions.
  Oxs_ArrayWrapper(const T&);
  T& operator=(const T&);

public:
  void Free(); // Calls destructors for objs, releases memory,
  /// and sets size to 0.

  void SetSize(UINT4m new_size);  // Throws and exception
  // if size!=0 on entry.  If you want to use this function,
  // call Free() first.

  UINT4m GetSize() const { return size; }

  Oxs_ArrayWrapper(UINT4m initial_size=0) : size(0), objs(NULL) {
    SetSize(initial_size);
  }

  ~Oxs_ArrayWrapper() { Free(); }

  T& operator[](UINT4m i) { return objs[i]; }
  const T& operator[](UINT4m i) const { return objs[i]; }
  /// NB: No range check!
};

template<class T>
void Oxs_ArrayWrapper<T>::Free()
{
  if(size==0 && objs==NULL) return; // Nothing to do
  if(size==0 || objs==NULL) {
    OXS_THROW(Oxs_ProgramLogicError,
	      "Oxs_ArrayWrapper<T>::Free():"
	      " Size and objs pointer incompatibility detected.");
  }
  size=0;
  delete[] objs;
  objs=NULL;
}

template<class T>
void Oxs_ArrayWrapper<T>::SetSize(UINT4m new_size)
{
  if(size!=0) {
    OXS_THROW(Oxs_ProgramLogicError,
	      "Oxs_ArrayWrapper<T>::SetSize():"
	      " Array not empty; call Free first.");
  }
  if(new_size==0) return; // Nothing to do.
  objs = new T[new_size];
  if(objs==NULL) {
    OXS_THROW(Oxs_NoMem,
	      "Oxs_ArrayWrapper<T>::SetSize():"
	      " Memory allocation error.");
  }
  size=new_size;
}

#endif // _OXS_UTIL

