/* FILE: atlasscalarfield.cc      -*-Mode: c++-*-
 *
 * Atlas scalar field object, derived from Oxs_ScalarField
 * class.
 *
 */

#include <string>
#include <vector>

#include "oc.h"
#include "nb.h"

#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "atlas.h"
#include "atlasscalarfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_AtlasScalarField);

/* End includes */

// Constructor
Oxs_AtlasScalarField::Oxs_AtlasScalarField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_ScalarField(name,newdtr,argstr)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas_obj);
  atlas_key.Set(atlas_obj.GetPtr());
  atlas_key.GetReadReference(); // Read lock held until *this is deleted.

  // Optional "default_value"
  OxsASFSpecifiedValue default_value;
  if(HasInitValue("default_value")) {
    default_value.value=GetRealInitValue("default_value");
    default_value.isset=1;
  }

  // Fill out value vector with default_value.  This is essentially
  // an empty padding if default.isset==0.
  unsigned int i;
  for(i=0;i<atlas_obj->GetRegionCount();i++) {
    values.push_back(default_value);
  }

  // Fill value vector with non-default values
  vector<string> vs;
  if(!FindInitValue("values",vs) || vs.size()<1) {
    string msg = "Missing specification for parameter \"values\"";
      throw Oxs_Ext::Error(this,msg.c_str());
  }
  if(vs.size()%2!=0) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"List associated with parameter \"values\" has"
		" %d entries, which is not an even number.",
		vs.size());
    throw Oxs_Ext::Error(this,buf);
  }
  for(i=0;i<vs.size();i+=2) {
    const string& name = vs[i];
    const string& value = vs[i+1];
    INT4m region_id = atlas_obj->GetRegionId(name);
    if(region_id<0) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Specified name, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  name.c_str(),atlas_obj->InstanceName());
      throw Oxs_Ext::Error(this,buf);
    }
    BOOL error;
    values[region_id].value = Nb_Atof(value.c_str(),error);
    if(error) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Value \"%s\" associated with region \"%s\""
		  " is not a valid numeric value.",
		  value.c_str(),name.c_str());
      throw Oxs_Ext::Error(this,buf);
    } else {
      values[region_id].isset = 1;
    }
  }
  DeleteInitValue("values");

  VerifyAllInitArgsUsed();
}

BOOL Oxs_AtlasScalarField::Init()
{ return 1; }

REAL8m
Oxs_AtlasScalarField::Value(const ThreeVector& pt) const
{
  const Oxs_Atlas* atlas = atlas_key.GetPtr();
  INT4m id = atlas->GetRegionId(pt);
  if(!values[id].isset) {
    char buf[1024];
    string regionname;
    atlas->GetRegionName(id,regionname);
    Oc_Snprintf(buf,sizeof(buf),
		"Error in function Value():"
		" input point (%g,%g,%g), in region %s of atlas %s,"
		" has not been assigned a value.",
		pt.x,pt.y,pt.z,
		regionname.c_str(),atlas->InstanceName());
    throw Oxs_Ext::Error(this,buf);
  }
  return values[id].value;
}
