/* FILE: cgevolve.h                 -*-Mode: c++-*-
 *
 * Concrete minimization evolver class, using conjugate gradients
 *
 */

#ifndef _OXS_CGEVOLVE
#define _OXS_CGEVOLVE

#include "minevolver.h"
#include "key.h"
#include "output.h"

/* End includes */

class Oxs_CGEvolve:public Oxs_MinEvolver {
private:
  UINT4m energy_calc_count;
  UINT4m cycle_count;
  UINT4m gradient_reset_count; // Re-initialize minimization
  /// direction from gradient (as opposed to conjugate gradient)
  /// each time cycle_count%gradient_reset_count == 0.

  void GetEnergyAndmxHxm
  (const Oxs_SimState* state,          // Import
   Oxs_MeshValue<REAL8m>* energy,      // Export
   Oxs_MeshValue<ThreeVector>* mxHxm); // Export

  void GetRelativeEnergyAndDerivative
  (const Oxs_SimState* state, // Import
   REAL8m time_offset,        // Import
   REAL8m &relenergy,         // Export
   REAL8m &derivative);        // Export

  // 1D minimizations take place along line basept + t*basedir.
  Oxs_ConstKey<Oxs_SimState> base_state;
  Oxs_MeshValue<ThreeVector> base_mxHxm;
  Oxs_MeshValue<ThreeVector> base_direction;
  Oxs_MeshValue<REAL8m> base_energy;
  REAL8m base_mxHxm_mag_sq;
  REAL8m base_direction_max_mag;

  // Cache space
  UINT4m cache_state_id;
  Oxs_MeshValue<REAL8m> cache_energy;
  Oxs_MeshValue<ThreeVector> cache_mxHxm;

  // Scratch space
  Oxs_MeshValue<REAL8m> scratch_energy;
  Oxs_MeshValue<ThreeVector> scratch_field;

  BOOL bracket_flag; // If true, then bracket is set.
  UINT4m bracket_count; // Number of bracket attempt calls,
  /// reset in Init() (i.e., on problem start).
  REAL8m minimum_bracket_step;
  REAL8m maximum_bracket_step;
  REAL8m scaled_minimum_bracket_step;
  REAL8m scaled_maximum_bracket_step;
  REAL8m bracket_start_step;
  struct Bracket {
    REAL8m offset;  // Location is basept + offset*basedir
    REAL8m E;       // Energy relative to base_state
    REAL8m Ep;      // Derivative of E in direction basedir.
    Bracket() : offset(0.), E(0.), Ep(0.) {}
  };
  Bracket leftpt,rightpt;

  void FillBracket(REAL8m offset,
		   const Oxs_SimState* oldstate,
		   Oxs_Key<Oxs_SimState>& statekey,
		   Bracket& endpt);
  BOOL FindBracketStep(const Oxs_SimState* oldstate,
		       Oxs_Key<Oxs_SimState>& statekey);
  /// Returns 1 if step taken, 0 otherwise

  BOOL line_minimum_flag; // True if line minimum detected
  Bracket minpt; // Set if line_minimum_flag is true
  UINT4m line_minimum_count; // Number of line minimization
  /// calls, reset in Init() (i.e., on problem start).
  REAL8m line_minimum_relwidth; // Line minimum considered
  /// found when relative size of bracketed region is smaller
  /// than this.
  REAL8m last_minimum_reduction_ratio;
  REAL8m next_to_last_minimum_reduction_ratio;
  BOOL FindLineMinimumStep(const Oxs_SimState* oldstate,
			   Oxs_Key<Oxs_SimState>& statekey);
  /// Returns 1 if step taken, 0 otherwise

  // Field outputs
  void UpdateDerivedFieldOutputs(const Oxs_SimState& state);
  Oxs_VectorFieldOutput<Oxs_CGEvolve> total_H_field_output;
  Oxs_VectorFieldOutput<Oxs_CGEvolve> mxHxm_output;

  // Scalar outputs
  void UpdateDerivedOutputs(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_CGEvolve> energy_calc_count_output;
  Oxs_ScalarOutput<Oxs_CGEvolve> cycle_count_output;
  Oxs_ScalarOutput<Oxs_CGEvolve> max_mxHxm_output;
  Oxs_ScalarOutput<Oxs_CGEvolve> delta_E_output;
  Oxs_ScalarOutput<Oxs_CGEvolve> total_energy_output;
  Oxs_ScalarOutput<Oxs_CGEvolve> bracket_count_output;
  Oxs_ScalarOutput<Oxs_CGEvolve> line_min_count_output;

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  virtual BOOL Init();
  Oxs_CGEvolve(const char* name,     // Child instance id
	       Oxs_Director* newdtr, // App director
	       const char* argstr);  // MIF input block parameters
  virtual ~Oxs_CGEvolve();

  virtual  BOOL
  Step(const Oxs_MinDriver* driver,
       Oxs_ConstKey<Oxs_SimState> current_state,
       Oxs_Key<Oxs_SimState>& next_state);
  // Returns true if step was successful, false if
  // unable to step as requested.
};

#endif // _OXS_CGEVOLVE
