/* FILE: exchange6ngbr.cc            -*-Mode: c++-*-
 *
 * 6 neighbor exchange energy on rectangular mesh,
 * derived from Oxs_Energy class.
 *
 */

#include "atlas.h"
#include "nb.h"
#include "key.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "exchange6ngbr.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_Exchange6Ngbr);

/* End includes */


// Constructor
Oxs_Exchange6Ngbr::Oxs_Exchange6Ngbr(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr),
    A_size(0), A(NULL), mesh_id(0)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas);
  atlaskey.Set(atlas.GetPtr());
  /// Dependency lock is held until *this is deleted.

  // Check for optional default_A parameter; default is 0.
  REAL8m default_A = GetRealInitValue("default_A",0.0);

  // Allocate A matrix.  Because raw pointers are used, a memory
  // leak will occur if an exception is thrown inside this constructor.
  A_size = atlas->GetRegionCount();
  if(A_size<1) {
    string msg=string("Oxs_Atlas object ")
      + atlas->InstanceName()
      + string(" must contain at least one region.");

    throw Oxs_Ext::Error(msg.c_str());
  }
  A = new REAL8m*[A_size];
  A[0] = new REAL8m[A_size*A_size];
  UINT4m i;
  for(i=1;i<A_size;i++) A[i] = A[i-1] + A_size;
  for(i=0;i<A_size*A_size;i++) A[0][i] = default_A;

  // Fill A matrix
  vector<string> params;
  FindRequiredInitValue("A",params);
  if(params.size()<1) {
    throw Oxs_Ext::Error(this,"Empty parameter list for key \"A\"");
  }
  if(params.size()%3!=0) {
      char buf[512];
      Oc_Snprintf(buf,sizeof(buf),
		  "Number of elements in A sub-list must be"
		  " divisible by 3"
		  " (actual sub-list size: %d)",
		  params.size());
      throw Oxs_Ext::Error(this,buf);
  }
  for(i=0;i<params.size();i+=3) {
    INT4m i1 = atlas->GetRegionId(params[i]);
    INT4m i2 = atlas->GetRegionId(params[i+1]);
    if(i1<0) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
		  "First entry in A[%d] sub-list, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  i/3,params[i].c_str(),atlas->InstanceName());
      throw Oxs_Ext::Error(this,buf);
    }
    if(i2<0) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
		  "Second entry in A[%d] sub-list, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  i/2,params[i+1].c_str(),atlas->InstanceName());
      throw Oxs_Ext::Error(this,buf);
    }
    BOOL err;
    REAL8m Apair = Nb_Atof(params[i+2].c_str(),err);
    if(err) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
		  "Third entry in A[%d] sub-list, \"%s\","
		  " is not a valid floating point number.",
		  i/3,params[i+2].c_str());
      throw Oxs_Ext::Error(this,buf);
    }
    A[i1][i2]=Apair;
    A[i2][i1]=Apair; // A should be symmetric
  }
  DeleteInitValue("A");

  VerifyAllInitArgsUsed();
}

Oxs_Exchange6Ngbr::~Oxs_Exchange6Ngbr()
{
  if(A_size>0 && A!=NULL) {
    delete[] A[0];
    delete[] A;
  }
}

void Oxs_Exchange6Ngbr::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  // See if mesh and/or atlas has changed.
  if(mesh_id !=  state.mesh->Id() || !atlaskey.SameState()) {
    // Setup region mapping
    mesh_id = 0; // Safety
    UINT4m size = state.mesh->Size();
    region_id.AdjustSize(state.mesh);
    ThreeVector location;
    for(UINT4m i=0;i<size;i++) {
      state.mesh->Center(i,location);
      if((region_id[i] = atlas->GetRegionId(location))<0) {
	string msg="Import mesh to Oxs_Exchange6Ngbr::GetEnergy()"
	  " routine of object " +  string(InstanceName())
	  + string(" has points outside atlas ")
	  + string(atlas->InstanceName());
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    mesh_id = state.mesh->Id();
  }
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    string msg="Import mesh to Oxs_Exchange6Ngbr::GetEnergy()"
      " routine of object " +  string(InstanceName())
      + " is not an Oxs_RectangularMesh object.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  REAL8m wgtx = -1.0/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = -1.0/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = -1.0/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  REAL8m hcoef = -2/MU0;

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address
	ThreeVector base = spin[i];
	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}
	REAL8m* Arow = A[region_id[i]];
	ThreeVector sum(0.,0.,0.);
	if(z>0) {
	  UINT4m j = i-xydim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtz*(spin[j] - base);
	}
	if(y>0) {
	  UINT4m j = i-xdim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgty*(spin[j] - base);
	}
	if(x>0) {
	  UINT4m j = i-1;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtx*(spin[j] - base);
	}
	if(x<xdim-1) {
	  UINT4m j = i+1;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtx*(spin[j] - base);
	}
	if(y<ydim-1) {
	  UINT4m j = i+xdim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgty*(spin[j] - base);
	}
	if(z<zdim-1) {
	  UINT4m j = i+xydim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtz*(spin[j]- base);
	}
	field[i] = (hcoef*Msii) * sum;
	energy[i] = (sum * base);
      }
    }
  }
}
