/* FILE: exchangeptwise.cc            -*-Mode: c++-*-
 *
 * 6 neighbor exchange energy on rectangular mesh, with
 * exchange coefficient A specified on a cell-by-cell
 * basis.  The effective coefficient between two neighboring
 * cells is calculated on the fly via 2.A_1.A_2/(A_1+A_2)
 *
 */

#include "atlas.h"
#include "nb.h"
#include "key.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "exchangeptwise.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ExchangePtwise);

/* End includes */


// Constructor
Oxs_ExchangePtwise::Oxs_ExchangePtwise(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr), mesh_id(0)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("A",Oxs_ScalarField,A_init);
  VerifyAllInitArgsUsed();
}

Oxs_ExchangePtwise::~Oxs_ExchangePtwise()
{}

void Oxs_ExchangePtwise::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    string msg="Import mesh to Oxs_ExchangePtwise::GetEnergy()"
      " routine of object " +  string(InstanceName())
      + " is not an Oxs_RectangularMesh object.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  if(mesh_id != state.mesh->Id()) {
    // This is either the first pass through, or else mesh
    // has changed.
    mesh_id = 0;
    A_init->FillMeshValue(state.mesh,A);
    // Check that we won't have any divide-by-zero problems
    // when we use A.
    for(UINT4m z=0;z<zdim;z++) {
      for(UINT4m y=0;y<ydim;y++) {
	for(UINT4m x=0;x<xdim;x++) {
	  UINT4m i = mesh->Index(x,y,z); // Get base linear address
	  REAL8m Ai = A[i];
	  for(UINT4m dir=0;dir<3;dir++) { // 0=>x, 1=>y, 2=>z
	    UINT4m j=0,x2=x,y2=y,z2=z;
	    switch(dir) {
	      case 0:  if(x+1<xdim) { j=i+1;     x2+=1; } break;
	      case 1:  if(y+1<ydim) { j=i+xdim;  y2+=1; } break;
	      default: if(z+1<zdim) { j=i+xydim; z2+=1; } break;
	    }
	    if(j==0) continue; // Edge case; skip
	    REAL8m Aj=A[j];
	    if(fabs(Ai+Aj)<1.0
	       && fabs(Ai*Aj)>(DBL_MAX/2.)*(fabs(Ai+Aj))) {
	      char buf[1024];
	      Oc_Snprintf(buf,sizeof(buf),
		 "Exchange coefficient A initialization in"
		 " Oxs_ExchangePtwise::GetEnergy()"
		 " routine of object %s yields infinite value"
		 " for A: A(%d,%d,%d)=%g, A(%d,%d,%d)=%g\n",
		 InstanceName(),x,y,z,Ai,x2,y2,z2,Aj);
	      throw Oxs_Ext::Error(buf);
	    }
	  }
	} // x++
      } // y++
    } // z++
    mesh_id = state.mesh->Id();
  } // mesh_id != state.mesh->Id()

  // First loop through and zero out entire energy and field
  // meshes.  This could be done inside the main body of the
  // loop, but it makes the logic a bit more complicated.
  UINT4m size = state.mesh->Size();
  UINT4m index;
  for(index=0;index<size;index++) energy[index]=0.;
  for(index=0;index<size;index++) field[index].Set(0.,0.,0.);

  // Now run through mesh and accumulate energies and fields.
  REAL8m wgtx = 1.0/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = 1.0/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = 1.0/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  REAL8m hcoef = 2/MU0;

  // Exchange energy density is
  //
  //     A.(1-m_i*m_j)/d^2                      (1)
  //
  // where A is effective exchange coefficient, and d is
  // the distance between the centers of cells i and j.  In
  // this code effective A is calculated from A_i and A_j
  // via A = 2.A_i.A_j/(A_i+A_j).  This energy density
  // is accumated at both cells i & j.
  //   The fields are
  //
  //     H_i = (2/MU0).[A/(Ms_i.d^2)].m_j       (2)
  //     H_j = (2/MU0).[A/(Ms_j.d^2)].m_i
  //
  // where the lead factor of 2 comes because in the total
  // energy expression, (1) occurs twice, once for cell i
  // and once for cell j.

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address
	ThreeVector spini = spin[i];
	REAL8m Msii = Ms_inverse[i];
	REAL8m Ai = A[i];
	if(Msii == 0.0 || Ai==0.0) continue;
	// Iterate through 3 neighbor directions, +x, +y and +z
	for(UINT4m dir=0;dir<3;dir++) { // 0=>x, 1=>y, 2=>z
	  UINT4m j=0;
	  REAL8m wgt=0.;
	  switch(dir) {
	  case 0:  if(x+1<xdim) { j=i+1;     wgt=wgtx; } break;
	  case 1:  if(y+1<ydim) { j=i+xdim;  wgt=wgty; } break;
	  default: if(z+1<zdim) { j=i+xydim; wgt=wgtz; } break;
	  }
	  if(j==0) continue; // Edge case; skip
	  REAL8m Aj=A[j];
	  REAL8m Msij=Ms_inverse[j];
	  if(Msij!=0.0 && Aj!=0.0) {
	    REAL8m acoef = 2*wgt*((Ai*Aj)/(Ai+Aj));
	    ThreeVector mdiff = spini - spin[j];
	    mdiff *= acoef;
	    REAL8m eij = mdiff*spini;
	    energy[i]+=eij;
	    energy[j]+=eij;
	    mdiff *= hcoef;
	    field[i] += -1*Msii*mdiff;
	    field[j] +=    Msij*mdiff;
	  }
	} // Neighbor direction loop
      } // x++
    } // y++
  } // z++
}

