/* FILE: randomsiteexchange.cc            -*-Mode: c++-*-
 *
 * 6 neighbor exchange energy on random links on rectangular mesh,
 * with random A.
 *
 */

#include "nb.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "randomsiteexchange.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_RandomSiteExchange);

/* End includes */


// Constructor
Oxs_RandomSiteExchange::Oxs_RandomSiteExchange(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr),
    mesh_id(0)
{
  // Process arguments
  Amin = GetRealInitValue("Amin");
  Amax = GetRealInitValue("Amax");
  if(Amax<Amin) {
    string msg="Invalid initialization detected for object "
      + string(InstanceName())
      + ": Amax must be >= Amin";
    throw Oxs_Ext::Error(msg.c_str());
  }
  linkprob = GetRealInitValue("linkprob");
  if(linkprob<0.) {
    string msg="Invalid initialization detected for object "
      + string(InstanceName())
      + ": linkprob value must be bigger than 0.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(linkprob>1.) {
    string msg="Invalid initialization detected for object "
      + string(InstanceName())
      + ": linkprob value must lie in the range [0,1].";
    throw Oxs_Ext::Error(msg.c_str());
  }

  VerifyAllInitArgsUsed();
}

Oxs_RandomSiteExchange::~Oxs_RandomSiteExchange()
{}

void Oxs_RandomSiteExchange::FillLinkList
(const Oxs_RectangularMesh* mesh) const
{
  // Iterate through mesh; at each cell roll iterate through
  // its 3 inferior neighbors.  For each link, roll dice to
  // see if that link gets a non-zero link energy.  If so,
  // add the two associated cells along with a random
  // exchange coefficient A value into the link list.
  links.clear();  // Throw away previous links, if any.

  if(linkprob<=0.) return; // Set no links

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;
  REAL8m Arange = Amax-Amin;

  REAL8m wgtx = 1/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = 1/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = 1/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	for(UINT4m li=0;li<3;li++) { // Link index: 0=x, 1=y, 2=z
	  REAL8m luck=Oc_UnifRand();
	  if(luck<=linkprob) {
	    // Make this link
	    UINT4m offset=0;
	    REAL8m wgt=0.;
	    switch(li) {
	      case 0:  if(x+1<xdim) offset=1;      wgt=wgtx;  break;
	      case 1:  if(y+1<ydim) offset=xdim;   wgt=wgty;  break;
	      default: if(z+1<zdim) offset=xydim;  wgt=wgtz;  break;
	    }
	    if(offset>0) {
	      // Link partner is also inside mesh
	      Oxs_RandomSiteExchangeLinkParams link;
	      link.index1 = mesh->Index(x,y,z); // Get base linear address
	      link.index2 = link.index1 + offset;
	      link.Acoef = (Amin + Arange*Oc_UnifRand())*wgt;
	      links.push_back(link);
	    }
	  }
	}
      }
    }
  }
}

void Oxs_RandomSiteExchange::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    string msg="Import mesh to Oxs_RandomSiteExchange::GetEnergy()"
      " routine of object " +  string(InstanceName())
      + " is not an Oxs_RectangularMesh object.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  // If mesh has changed, re-pick link selections
  if(mesh_id !=  mesh->Id()) {
    FillLinkList(mesh);
    mesh_id=mesh->Id();
  }

  // Zero entire energy and field meshes
  UINT4m size=mesh->Size();
  UINT4m index;
  for(index=0;index<size;index++) energy[index]=0.;
  for(index=0;index<size;index++) field[index].Set(0.,0.,0.);

  // Iterate through link list and accumulate energies and fields
  REAL8m hcoef = 2.0/MU0;
  vector<Oxs_RandomSiteExchangeLinkParams>::const_iterator it
    = links.begin();
  for(it=links.begin();it!=links.end();++it) {
    UINT4m i = it->index1;
    UINT4m j = it->index2;
    if(Ms_inverse[i]==0. || Ms_inverse[j]==0.) continue; // Ms=0; skip
    REAL8m acoef = it->Acoef;
    ThreeVector mdiff = spin[i]-spin[j];
    mdiff *= acoef;
    REAL8m elink = mdiff*spin[i]; // Energy
    energy[i] += elink;
    energy[j] += elink;
    mdiff *= hcoef;
    field[i] += -1*Ms_inverse[i]*mdiff;
    field[j] +=    Ms_inverse[j]*mdiff;
  }
}
