/* FILE: randomvectorfield.cc      -*-Mode: c++-*-
 *
 * Random vector field object, derived from Oxs_VectorField class.
 *
 */

#include "oc.h"
#include "nb.h"

#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "randomvectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_RandomVectorField);

/* End includes */


// Constructor
Oxs_RandomVectorField::Oxs_RandomVectorField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorField(name,newdtr,argstr)
{
  // Process arguments.
  min_norm=GetRealInitValue("min_norm");
  max_norm=GetRealInitValue("max_norm");

  VerifyAllInitArgsUsed();

  if(min_norm>max_norm) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Requested min_norm value (%g) bigger than max_norm (%g)",
		min_norm,max_norm);
    throw Oxs_Ext::Error(this,buf);
  }
}

BOOL Oxs_RandomVectorField::Init()
{
  return 1;
}

void
Oxs_RandomVectorField::Value
(const ThreeVector& /* pt */,
 ThreeVector& value) const
{
  if(max_norm==min_norm) {
    // No spread in magnitudes
    value.Random(max_norm);
  } else {
    // Vary both direction and magnitude
    REAL8m mag = (max_norm-min_norm)*Oc_UnifRand()+min_norm;
    value.Random(mag);
  }
}

void
Oxs_RandomVectorField::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<ThreeVector>& array) const
{
  array.AdjustSize(mesh);
  UINT4m i,size=mesh->Size();
  if(max_norm==min_norm) {
    // No spread in magnitudes
    for(i=0;i<size;i++) {
      array[i].Random(max_norm);
    }
  } else {
    // Vary both direction and magnitude
    REAL8m norm_range=max_norm-min_norm;
    for(i=0;i<size;i++) {
      REAL8m mag = norm_range*Oc_UnifRand()+min_norm;
      array[i].Random(mag);
    }
  }
}

void
Oxs_RandomVectorField::FillMeshValueOrthogonal
(const Oxs_Mesh* mesh,
 const Oxs_MeshValue<ThreeVector>& orth_array,
 Oxs_MeshValue<ThreeVector>& array) const
{
  if(!orth_array.CheckMesh(mesh)) {
    string msg=
      string("Mesh size incompatibility detected in"
	     " Oxs_RandomVectorField::FillMeshValueOrthogonal()"
	     " method of instance ")
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  const REAL8m norm_range=max_norm-min_norm;
  REAL8m dot=0.0;
  ThreeVector vec,tempvec;
  UINT4m i,size=mesh->Size();
  array.AdjustSize(mesh);
  for(i=0;i<size;i++) {
    // First produce vector perpendicular to orth_array[i]
    tempvec=orth_array[i];
    tempvec.SetMag(1.0);
    int j=0;
    while(1) {
      vec.Random(1.0);
      dot = tempvec*vec;
      if(fabs(dot)<0.99) break;
      if(++j>32) {
	string msg=
	  string("Programming error?  Unable to obtain vector"
		 " in specified plane in"
		 " Oxs_RandomVectorField::FillMeshValueOrthogonal()"
		 " method of instance ")
	  + string(InstanceName());
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    vec -= dot*tempvec;

    // Adjust to proper size
    REAL8m mag = min_norm;
    if(norm_range!=0.0) {
      // Vary magnitude
      mag += norm_range*Oc_UnifRand();
    }
    vec.SetMag(mag);
    array[i] = vec;
  }
}




