/* FILE: rectangularmesh.cc         -*-Mode: c++-*-
 *
 * Rectangular mesh, derived from Oxs_Mesh class.
 *
 */

#include "vf.h"
#include "atlas.h"
#include "meshvalue.h"
#include "rectangularmesh.h"
#include "director.h"
#include "nb.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy
#include "util.h"


// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_RectangularMesh);

/* End includes */

/////////////////////////////////////////////////////////////////////
// Oxs_RectangularMesh
Oxs_RectangularMesh::Oxs_RectangularMesh
(const char* name,     // Child instance id
 Oxs_Director* newdtr, // App director
 const char* argstr)   // MIF input block parameters
  : Oxs_Mesh(name,newdtr,argstr)
{
  // Process arguments
  cellsize = GetThreeVectorInitValue("cellsize");
  
  Oxs_OwnedPointer<Oxs_Atlas> atlas;
  OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas);

  if(cellsize.x<=0.0 || cellsize.y<=0.0 || cellsize.z<=0.0) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  VerifyAllInitArgsUsed();

  cellvolume=cellsize.x*cellsize.y*cellsize.z;

  Oxs_Box box;
  atlas->GetWorldExtents(box);
  REAL8m xrange = box.GetMaxX() - box.GetMinX();
  REAL8m yrange = box.GetMaxY() - box.GetMinY();
  REAL8m zrange = box.GetMaxZ() - box.GetMinZ();
  if(xrange<=0. || yrange<=0. || zrange<=0.) {
    string msg="Invalid atlas range detected for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  base = ThreeVector(box.GetMinX()+cellsize.x/2.,
                     box.GetMinY()+cellsize.y/2.,
                     box.GetMinZ()+cellsize.z/2.);

  xdim = static_cast<UINT4m>(OC_ROUND(xrange/cellsize.x));
  ydim = static_cast<UINT4m>(OC_ROUND(yrange/cellsize.y));
  zdim = static_cast<UINT4m>(OC_ROUND(zrange/cellsize.z));
  if(xdim<1 || ydim<1 || zdim<1) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + string("; minimum range smaller than cell dimension.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  xydim = xdim*ydim;
  elementcount = xydim*zdim;
  if(elementcount<xdim || elementcount<ydim || elementcount<zdim) {
    /// The above test is not complete, but will catch
    /// some overflows.
    string msg="UINT4m overflow: mesh size too large in "
      "Oxs_RectangularMesh constructor";
    throw Oxs_Ext::Error(msg.c_str());
  }

  if( fabs(xdim*cellsize.x - xrange) > 0.01*cellsize.x ||
      fabs(ydim*cellsize.y - yrange) > 0.01*cellsize.y ||
      fabs(zdim*cellsize.z - zrange) > 0.01*cellsize.z   ) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": range is not an integral multiple of cellsize.";
    throw Oxs_Ext::Error(msg.c_str());
  }
}


void Oxs_RectangularMesh::GetBoundingBox(Oxs_Box& bbox) const
{
  bbox.Set(base.x-cellsize.x/2., base.x-cellsize.x/2.+xdim*cellsize.x,
           base.y-cellsize.y/2., base.y-cellsize.y/2.+ydim*cellsize.y,
           base.z-cellsize.z/2., base.z-cellsize.z/2.+zdim*cellsize.z);
}

void Oxs_RectangularMesh::Center(UINT4m index,ThreeVector &location) const
{
#ifndef NDEBUG
  if(index>elementcount) {
    string msg="Index out of range "
      "(Oxs_RectangularMesh::Location(INT4m) const)";
    throw Oxs_Ext::Error(msg.c_str());
  }
#endif
  INT4m iz = index/(xdim*ydim);  index -= iz*xdim*ydim;
  INT4m iy = index/xdim;         index -= iy*xdim;
  INT4m ix = index;
  location.Set(base.x+ix*cellsize.x,
               base.y+iy*cellsize.y,
               base.z+iz*cellsize.z);
  return;
}


BOOL
Oxs_RectangularMesh::GetNeighborPoint
(const ThreeVector& pt,
 UINT4m ngbr_index,
 ThreeVector& ngbr_pt) const
{ // Fills ngbr_pt with location of neighbor element indexed
  // by "ngbr_index", relative to pt.  Returns 1 if
  // ngbr_pt < number of neighbors (currently 6); otherwise
  // 0 is returned, in which case ngbr_pt is unchanged.
  // NB: ngbr_index is 0-based.
  if(ngbr_index>5) return 0;
  int sign = 1 - 2*(ngbr_index%2); // 0,2,4 => +1, 1,3,5 => -1
  ngbr_pt = pt;
  switch(ngbr_index/2) {
    case 0:  ngbr_pt.x += sign*cellsize.x; break;
    case 1:  ngbr_pt.y += sign*cellsize.y; break;
    default: ngbr_pt.z += sign*cellsize.z; break;
  }
  return 1;
}

UINT4m
Oxs_RectangularMesh::BoundaryList
(const Oxs_Atlas& atlas,
 const string& region_name,
 const Oxs_ScalarField& bdry_surface,
 REAL8m bdry_value,
 const string& bdry_side,
 vector<UINT4m> &BoundaryIndexList) const
{ // Boundary extraction.  Returns list (technically, an STL vector) of
  // indices for those elements inside base_region that have a neighbor
  // (in the 6 nearest ngbr sense) such that the first element lies on
  // one side (the "inside") of the surface specified by the
  // Oxs_ScalarField bdry_surface + bdry_value, and the neighbor lies on
  // the other (the "outside").  If the bdry_side argument is "-", then
  // the "inside" of the surface is the set of those points x for which
  // bdry_surface.Value(x)<bdry_value.  If bdry_side is "+", then the
  // "inside" of the surface is the set of those points x for which
  // bdry_surface.Value(x)>bdry_value.
  // Return value is the number of entries in the export list.
  // NB: The tested neighbors may lie outside the mesh proper, allowing
  // elements on the edge of the mesh to be specified.
  BoundaryIndexList.clear();

  Oxs_Box world_box,work_box;
  GetBoundingBox(world_box);
  INT4m region_id = atlas.GetRegionId(region_name);
  if(region_id<0) {
    string msg=string("Region name ")
      + region_name
      + string(" not recognized by atlas ")
      + string(atlas.InstanceName())
      + string(" (Oxs_RectangularMesh::BoundaryList() in object")
      + string(InstanceName())
      + string(")");
    throw Oxs_Ext::Error(msg.c_str());
  }
  atlas.GetRegionExtents(region_id,work_box);
  work_box.Intersect(world_box); // work_box contains extents of
  /// base_region contained inside mesh
  UINT4m ixstart = UINT4m(ceil((work_box.GetMinX()-base.x)/cellsize.x));
  UINT4m iystart = UINT4m(ceil((work_box.GetMinY()-base.y)/cellsize.y));
  UINT4m izstart = UINT4m(ceil((work_box.GetMinZ()-base.z)/cellsize.z));
  UINT4m ixstop = UINT4m(floor((work_box.GetMaxX()-base.x)/cellsize.x))+1;
  UINT4m iystop = UINT4m(floor((work_box.GetMaxY()-base.y)/cellsize.y))+1;
  UINT4m izstop = UINT4m(floor((work_box.GetMaxZ()-base.z)/cellsize.z))+1;

  // Check sign
  int check_sign = 0;
  if(bdry_side.compare("+")==0)       check_sign =  1;
  else if(bdry_side.compare("-")==0)  check_sign = -1;
  else {
    string msg=string("Invalid boundary side representation: ")
      + bdry_side
      + string("  Should be either - or +.")
      + string(" (Oxs_RectangularMesh::BoundaryList() in object")
      + string(InstanceName())
      + string(")");
    throw Oxs_Ext::Error(msg.c_str());
  }

  UINT4m ixsize = ixstop-ixstart;
  for(UINT4m iz=izstart;iz<izstop;iz++) {
    for(UINT4m iy=iystart;iy<iystop;iy++) {
      UINT4m row_index = Index(ixstart,iy,iz);
      ThreeVector pt;
      Center(row_index,pt);
      for(UINT4m i=0; i<ixsize; ++i,pt.x+=cellsize.x) {
	if(check_sign*(bdry_surface.Value(pt)-bdry_value)<0) {
	  continue; // basept on wrong side of boundary
	}
	UINT4m ngbr_index=0;
	ThreeVector ngbr_pt;
	while(GetNeighborPoint(pt,ngbr_index++,ngbr_pt)) {
	  if(check_sign*(bdry_surface.Value(ngbr_pt)-bdry_value)<0) {
	    // Neighbor on "other" side of boundary
	    BoundaryIndexList.push_back(row_index+i);
	  }
	}
      }
    }
  }
  return BoundaryIndexList.size();
}



// File output.  Throws an exception on error.
void Oxs_RectangularMesh::WriteOvf
(Tcl_Channel channel,   // Name of output file
 const char* title,     // Long filename or title
 const char* desc,      // Description to embed in output file
 const char* valueunit, // Field units, such as "A/m".
 const char* meshtype,  // Either "rectangular" or "irregular"
 const char* datatype,  // Either "binary" or "text"
 const char* precision, // For binary, "4" or "8";
                        ///  for text, a printf-style format
 const Oxs_MeshValue<ThreeVector>* vec,  // Vector array
 const Oxs_MeshValue<REAL8m>* scale      // Optional scaling for vec
 /// Set scale to NULL to use vec values directly.
 ) const
{
  // Use default file writer for irregular mesh.
  if(strcmp("irregular",meshtype)==0) {
    ThreeVector stephints(cellsize.x,cellsize.y,cellsize.z);
    WriteOvfIrregular(channel,title,desc,valueunit,
		      datatype,precision,
		      vec,scale,&stephints);
    return;
  }

  if(strcmp("rectangular",meshtype)!=0) {
    string msg=string("Unrecognized mesh type request: ")
      + string(meshtype)
      + string("(Oxs_RectangularMesh::WriteOvf() in object")
      + string(InstanceName()) + string(")");
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Rectangular Mesh ////////////////////////////
  // Check import validity
  enum DataType { BINARY, TEXT };
  DataType dt = BINARY;
  if(strcmp("text",datatype)==0) {
    dt = TEXT;
  } else if(strcmp("binary",datatype)!=0){
    string errmsg = string("Bad datatype: \"") + string(datatype)
      + string("\"  Should be either \"binary\" or \"text\"");
    throw Oxs_Ext::Error(this,errmsg.c_str());
  }

  int datawidth = 0;
  string dataformat;
  if(dt == BINARY) {
    datawidth = atoi(precision);
    if(datawidth != 4 && datawidth != 8) {
      string errmsg = string("Bad precision: ") + string(precision)
	+ string("  Should be either 4 or 8.");
      throw Oxs_Ext::Error(this,errmsg.c_str());
    }
  } else {
    if(precision==NULL || precision[0]=='\0')
      precision="%# .17g";  // Default format
    string temp = string(precision) + string(" ");
    dataformat = temp + temp + string(precision) + string("\n");
  }

  if(!vec->CheckMesh(this)) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),"Size mismatch; input data length=%u,"
		" which is different than mesh size=%u",
		vec->Size(),Size());
    throw Oxs_Ext::Error(this,buf);
  }

  if(scale!=NULL && !scale->CheckMesh(this)) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),"Size mismatch; input scale data length=%u,"
		" which is different than mesh size=%u",
		scale->Size(),Size());
    throw Oxs_Ext::Error(this,buf);
  }

  // Write header
  Nb_FprintfChannel(channel,NULL,1024,"# OOMMF: rectangular mesh v1.0\n");
  Nb_FprintfChannel(channel,NULL,1024,"# Segment count: 1\n");
  Nb_FprintfChannel(channel,NULL,1024,"# Begin: Segment\n");
  Nb_FprintfChannel(channel,NULL,1024,"# Begin: Header\n");
  if(title==NULL || title[0]=='\0')
    Nb_FprintfChannel(channel,NULL,1024,"# Title: unknown\n");
  else
    Nb_FprintfChannel(channel,NULL,1024,"# Title: %s\n",title);
  if(desc!=NULL && desc[0]!='\0') {
    // Print out description, breaking at newlines as necessary.
    // Note: This block is optional
    const char *cptr1,*cptr2;
    cptr1=desc;
    while((cptr2=strchr(cptr1,'\n'))!=NULL) {
      Nb_FprintfChannel(channel,NULL,1024,"# Desc: %.*s\n",
                        (int)(cptr2-cptr1),cptr1);
      cptr1=cptr2+1;
    }
    if(*cptr1!='\0')
      Nb_FprintfChannel(channel,NULL,1024,"# Desc: %s\n",cptr1);
  }
  Nb_FprintfChannel(channel,NULL,1024,"# meshunit: m\n");
  Nb_FprintfChannel(channel,NULL,1024,"# valueunit: %s\n",valueunit);
  Nb_FprintfChannel(channel,NULL,1024,"# valuemultiplier: 1.0\n");

  Oxs_Box bbox;
  GetBoundingBox(bbox);
  Nb_FprintfChannel(channel,NULL,1024,
          "# xmin: %.16g\n# ymin: %.16g\n# zmin: %.16g\n"
          "# xmax: %.16g\n# ymax: %.16g\n# zmax: %.16g\n",
          bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ(),
          bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());

  // As of 6/2001, mmDisp supports display of out-of-plane rotations;
  // Representing the boundary under these conditions is awkward with
  // a single polygon.  So don't write the boundary line, and rely
  // on defaults based on the data range and step size.

  REAL8m minmag=0,maxmag=0;
  if(Size()>0) {
    minmag=DBL_MAX;
    for(UINT4m i=0;i<Size();i++) {
      REAL8m val=(*vec)[i].MagSq();
      if(scale!=NULL) {
        REAL8m tempscale=(*scale)[i];
        val*=tempscale*tempscale;
      }
      if(val<minmag && val>0) minmag=val; // minmag is smallest non-zero
      if(val>maxmag) maxmag=val;         /// magnitude.
    }
    if(minmag>maxmag) minmag=maxmag;
    maxmag=sqrt(maxmag);
    minmag=sqrt(minmag);
    minmag*=0.999;  // Underestimate lower bound by 0.1% to protect against
    /// rounding errors.
  }

  Nb_FprintfChannel(channel,NULL,1024,
                    "# ValueRangeMaxMag: %.16g\n",maxmag);
  Nb_FprintfChannel(channel,NULL,1024,
                    "# ValueRangeMinMag: %.16g\n",minmag);

  Nb_FprintfChannel(channel,NULL,1024,"# meshtype: rectangular\n");
  Nb_FprintfChannel(channel,NULL,1024,
          "# xbase: %.16g\n# ybase: %.16g\n# zbase: %.16g\n",
          base.x,base.y,base.z);
  Nb_FprintfChannel(channel,NULL,1024,
          "# xstepsize: %.16g\n# ystepsize: %.16g\n# zstepsize: %.16g\n",
          cellsize.x,cellsize.y,cellsize.z);
  Nb_FprintfChannel(channel,NULL,1024,
          "# xnodes: %d\n# ynodes: %d\n# znodes: %d\n",
          xdim,ydim,zdim);
  Nb_FprintfChannel(channel,NULL,1024,"# End: Header\n");

  // Write data block
  if(dt == BINARY) {
    if(datawidth==4) {
      Nb_FprintfChannel(channel,NULL,1024,"# Begin: Data Binary 4\n");
      REAL4 buf[3];
      buf[0]=1234567.;  // 4-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
        const ThreeVector& v = (*vec)[i];
        if(scale==NULL) {
          buf[0] = REAL4(v.x);
          buf[1] = REAL4(v.y);
          buf[2] = REAL4(v.z);
        } else {
          REAL8m tempscale=(*scale)[i];
          buf[0] = REAL4(tempscale*v.x);
          buf[1] = REAL4(tempscale*v.y);
          buf[2] = REAL4(tempscale*v.z);
        }
        Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,3);
        // Vf_OvfFileFormatSpecs::WriteBinary performs
        // byte-swapping as needed.
      }
      Nb_FprintfChannel(channel,NULL,1024,"\n# End: Data Binary 4\n");
    } else {
      Nb_FprintfChannel(channel,NULL,1024,"# Begin: Data Binary 8\n");
      REAL8 buf[3];
      buf[0]=123456789012345.;  // 8-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
        const ThreeVector& v = (*vec)[i];
        if(scale==NULL) {
          buf[0] = REAL4(v.x);
          buf[1] = REAL4(v.y);
          buf[2] = REAL4(v.z);
        } else {
          REAL8m tempscale=(*scale)[i];
          buf[0] = REAL4(tempscale*v.x);
          buf[1] = REAL4(tempscale*v.y);
          buf[2] = REAL4(tempscale*v.z);
        }
        Vf_OvfFileFormatSpecs::WriteBinary(channel,buf,3);
        // Vf_OvfFileFormatSpecs::WriteBinary performs
        // byte-swapping as needed.
      }
      Nb_FprintfChannel(channel,NULL,1024,"\n# End: Data Binary 8\n");
    }
  } else {
    Nb_FprintfChannel(channel,NULL,1024,"# Begin: Data Text\n");
    INT4m size=Size();
    for(INT4m i=0 ; i<size ; ++i) {
        const ThreeVector& v = (*vec)[i];
        if(scale==NULL) {
          Nb_FprintfChannel(channel,NULL,1024,dataformat.c_str(),
			    v.x,v.y,v.z);
        } else {
          REAL8m tempscale=(*scale)[i];
          Nb_FprintfChannel(channel,NULL,1024,dataformat.c_str(),
                  tempscale*v.x,tempscale*v.y,tempscale*v.z);
        }
    }
    Nb_FprintfChannel(channel,NULL,1024,"# End: Data Text\n");
  }
  Nb_FprintfChannel(channel,NULL,1024,"# End: Segment\n");
}


//////////////////////////////////////////////////////////////////////
// Volume summing routines.  This have advantage over the generic
// in that for Oxs_RectangularMesh all cells have same volume.
REAL8m Oxs_RectangularMesh::VolumeSum
(const Oxs_MeshValue<REAL8m>& scalar) const
{
  if(!scalar.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible scalar array import to"
			 " VolumeSum(const Oxs_MeshValue<REAL8m>&)");
  }
  const UINT4m size=Size();
  REAL8m sum=0.0;
  for(UINT4m i=0;i<size;i++) sum += scalar[i];
  sum *= cellvolume;
  return sum;
}

ThreeVector Oxs_RectangularMesh::VolumeSum
(const Oxs_MeshValue<ThreeVector>& vec) const
{
  if(!vec.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSum(const Oxs_MeshValue<ThreeVector>&)");
  }
  const UINT4m size=Size();
  ThreeVector sum(0.,0.,0.);
  for(UINT4m i=0;i<size;i++) sum += vec[i] ;
  sum *= cellvolume;
  return sum;
}

ThreeVector Oxs_RectangularMesh::VolumeSum
(const Oxs_MeshValue<ThreeVector>& vec,
 const Oxs_MeshValue<REAL8m>& scale
 ) const
{
  if(!vec.CheckMesh(this) || !scale.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSum(const Oxs_MeshValue<ThreeVector>&,"
			 " const Oxs_MeshValue<REAL8m>& scale)");
  }
  const UINT4m size=Size();
  ThreeVector sum(0.,0.,0.);
  for(UINT4m i=0;i<size;i++) {
    sum +=  scale[i] * vec[i];
  }
  sum *= cellvolume;
  return sum;
}


REAL8m Oxs_RectangularMesh::VolumeSumXp
(const Oxs_MeshValue<REAL8m>& scalar) const
{
  if(!scalar.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible scalar array import to"
			 " VolumeSumXp(const Oxs_MeshValue<REAL8m>&)");
  }

  Nb_Xpfloat sum=0.;
  const UINT4m size=Size();
  for(UINT4m i=0;i<size;i++) sum += scalar[i];
  sum *= cellvolume;
  return sum.GetValue();
}

ThreeVector Oxs_RectangularMesh::VolumeSumXp
(const Oxs_MeshValue<ThreeVector>& vec) const
{
  if(!vec.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSumXp(const Oxs_MeshValue<ThreeVector>&)");
  }

  Nb_Xpfloat sum_x=0.,sum_y=0.,sum_z=0.;
  const UINT4m size=Size();
  for(UINT4m i=0;i<size;i++) {
    sum_x += vec[i].x;
    sum_y += vec[i].y;
    sum_z += vec[i].z;
  }
  sum_x *= cellvolume;
  sum_y *= cellvolume;
  sum_z *= cellvolume;
  return ThreeVector(sum_x.GetValue(),sum_y.GetValue(),sum_z.GetValue());
}

ThreeVector Oxs_RectangularMesh::VolumeSumXp
(const Oxs_MeshValue<ThreeVector>& vec,
 const Oxs_MeshValue<REAL8m>& scale
 ) const
{
  if(!vec.CheckMesh(this) || !scale.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " VolumeSumXp(const Oxs_MeshValue<ThreeVector>&,"
			 " const Oxs_MeshValue<REAL8m>& scale)");
  }

  Nb_Xpfloat sum_x=0.,sum_y=0.,sum_z=0.;
  const UINT4m size=Size();
  for(UINT4m i=0;i<size;i++) {
    REAL8m wgt = scale[i];
    sum_x += wgt*vec[i].x;
    sum_y += wgt*vec[i].y;
    sum_z += wgt*vec[i].z;
  }
  sum_x *= cellvolume;
  sum_y *= cellvolume;
  sum_z *= cellvolume;
  return ThreeVector(sum_x.GetValue(),sum_y.GetValue(),sum_z.GetValue());
}


////////////////////////////////////////////////////////////////////////
// Max angle routine.  This routine returns the minimum dot
// product between neighboring vectors across the mesh, for all
// those vectors for which the corresponding entry in zero_check
// is non-zero.
//   If the input Oxs_MeshValue<ThreeVector> array contents are all
// unit vectors, then the max angle can be construed from acos(mindot).
//   NB: These routines are likely sensitive to boundary conditions.
REAL8m Oxs_RectangularMesh::MinNeighborDot
(const Oxs_MeshValue<ThreeVector>& vec,
 const Oxs_MeshValue<REAL8m>& zero_check) const
{
  if(!vec.CheckMesh(this)) {
    throw Oxs_Ext::Error(this,
			 "Incompatible import array to"
			 " MinNeighborDot(const Oxs_MeshValue<ThreeVector>&,"
			 " const Oxs_MeshValue<REAL8m>& scale)");
  }
  if(elementcount<1) return 0.0;
  if(elementcount<2) return vec[0].MagSq(); // Not certain what most
  /// sensible return is in 1 cell case.  This seems as good as any.
  /// If the input is suppose to be unit spins, then the caller might
  /// want to handle the 1 cell case itself.

  REAL8m mindot1 = DBL_MAX;
  REAL8m mindot2 = DBL_MAX;

  // Process in 3 passes.  First pass compare neighbors along
  // x-axis direction, 2nd do neighbors along y-axis direction,
  // and last do neighbors along z-axis direction.  On a machine
  // with good cache control, all 3 passes take similar time.
  //   The code below tries to store one of the ThreeVectors in
  // cpu registers to cut down on memory loads.  This is done by
  // comparing forward and backwards and then hopping by 2.  This
  // technique appears to give some efficiency gains; this has not
  // been widely checked, however, and one expects it to be a win
  // only on machines that are floating-point register rich.  If
  // successful, this method cuts down memory loads from 6n
  // to 4.5n, where n is the number of cells.  It is possible to
  // cut this down farther to 3.5n by using a 1-pass checkerboard
  // access pattern, but the edge cases look to make a messy
  // implementation.
  //   Another approach to speed up this computation is to prune
  // some of the dot products by using a triangle inequality.
  // For example, the angle between spins A and C will not be more
  // than the sum of the angles A to B and B to C.  So, given
  // A to B and B to C, one may be able to infer that A to C
  // is smaller than current max ang without explicit calculation.
  // Whether this approach could be done in an efficient enough
  // manner to beat the exhaustive search method remains to be
  // seen.  Also not clear to me if something like a triangle
  // inequality holds for arbitrary vectors, or only for case
  // where all spins are the same size.

  // First check min dot's for x-neighbors
  if(xdim>1) {
    UINT4m x,index=0;
    UINT4m row,row_count = ydim*zdim;
    for(row=0;row<row_count;row++) {
      ++index; // Skip past first cell in each row
      for(x=1;x<xdim-1;x+=2) {
	// Here x-1, x and x+1 are always valid indices
	if(zero_check[index]!=0.0) {
	  ThreeVector temp = vec[index];
	  REAL8m dot1 = temp*vec[index-1];
	  REAL8m dot2 = temp*vec[index+1];
	  if(zero_check[index-1]!=0.0 && dot1<mindot1) mindot1=dot1;
	  if(zero_check[index+1]!=0.0 && dot2<mindot2) mindot2=dot2;
	}
	index+=2;
      }
      if(xdim%2==0) { // If xdim is even, then there are an odd
	/// number of neighbor pairs in a row, so there will be
	/// one left over from the above x-loop.
	if(zero_check[index]!=0.0 && zero_check[index-1]!=0.0) {
	  ThreeVector temp = vec[index];
	  REAL8m dot1 = temp*vec[index-1];
	  if(dot1<mindot1) mindot1=dot1;
	}
	++index; // Wrap around to next row.
      }
    }
  }

  // Next check min dot's for y-neighbors
  if(ydim>1) {
    UINT4m x,y,z,index=0;
    for(z=0;z<zdim;z++) {
      index+=xdim; // Skip over first row.
      for(y=1;y<ydim-1;y+=2) {
	// Here y-1, y and y+1 are always valid indices
	for(x=0;x<xdim;x++) {
	  if(zero_check[index]!=0.0) {
	    ThreeVector temp = vec[index];
	    REAL8m dot1 = temp*vec[index-xdim];
	    REAL8m dot2 = temp*vec[index+xdim];
	    if(zero_check[index-xdim]!=0.0 && dot1<mindot1) mindot1=dot1;
	    if(zero_check[index+xdim]!=0.0 && dot2<mindot2) mindot2=dot2;
	  }
	  index++;
	}
	index+=xdim; // Jump over already processed row.
      }
      if(ydim%2==0) { // If ydim is even, then there are an odd
	/// number of neighbor pairs in a y-column, so there will
	/// be one row left over after the above y-loop.
	for(x=0;x<xdim;x++) {
	  if(zero_check[index]!=0.0 && zero_check[index-xdim]!=0.0) {
	    REAL8m dot1 = vec[index]*vec[index-xdim];
	    if(dot1<mindot1) mindot1=dot1;
	  }
	  index++;
	}
      }
    }
  }

  // Finally check min dot's for z-neighbors
  if(zdim>1) {
    UINT4m p,z,index=0;
    for(z=1;z<zdim-1;z+=2) {
      index += xydim; // Jump over one plane
      // Here z-1, z and z+1 are always valid indices
      for(p=0;p<xydim;p++) {
	if(zero_check[index]!=0.0) {
	  ThreeVector temp = vec[index];
	  REAL8m dot1 = temp*vec[index-xydim];
	  REAL8m dot2 = temp*vec[index+xydim];
	  if(zero_check[index-xydim]!=0.0 && dot1<mindot1) mindot1=dot1;
	  if(zero_check[index+xydim]!=0.0 && dot2<mindot2) mindot2=dot2;
	}
	++index;
      }
    }
    if(zdim%2==0) {
      // Process top two layers.
      // Note: Because xydim isn't added to index after
      //   last pass through above loop, index points to
      //   layer zdim-2 instead of zdim-1; thus the proper
      //   indices are "index" and "index+xydim", *not*
      //   "index-xydim".
      for(p=0;p<xydim;p++) {
	if(zero_check[index]!=0.0 && zero_check[index+xydim]!=0.0) {
	  REAL8m dot2 = vec[index]*vec[index+xydim];
	  if(dot2<mindot2) mindot2=dot2;
	}
	++index;
      }
    }
  }

  REAL8m mindot = mindot1;
  if(mindot2<mindot) mindot=mindot2;

  if(mindot == DBL_MAX) mindot=0.0; // No valid (non-zero) vec's.

  return mindot;
}
