/* FILE: scriptscalarfield.cc      -*-Mode: c++-*-
 *
 * Tcl script scalar field object, derived from Oxs_ScalarField
 * class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "atlas.h"
#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "scriptscalarfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ScriptScalarField);

/* End includes */

// Constructor
Oxs_ScriptScalarField::Oxs_ScriptScalarField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_ScalarField(name,newdtr,argstr)
{
  // Determine bounding box.
  Oxs_Box bbox;
  if(HasInitValue("atlas")) {
    // Get bounding box from specified atlas.
    Oxs_OwnedPointer<Oxs_Atlas> atlas;
    OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas);
    atlas->GetWorldExtents(bbox);
  } else {
    // Atlas not specified, so check for and use x/y/zrange data.
    if(!HasInitValue("xrange") || !HasInitValue("yrange")
       || !HasInitValue("zrange")) {
      throw Oxs_Ext::Error(this,
			   "Neither atlas nor x/y/zrange data specified.");
    }
    vector<REAL8m> xrange,yrange,zrange;
    GetRealVectorInitValue("xrange",2,xrange);
    GetRealVectorInitValue("yrange",2,yrange);
    GetRealVectorInitValue("zrange",2,zrange);
    if(!bbox.CheckOrderSet(xrange[0],xrange[1],
			   yrange[0],yrange[1],
			   zrange[0],zrange[1])) {
      throw Oxs_Ext::Error(this,"Bad order in x/y/zrange data.");
    }
  }

  basept.Set(bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ());
  ThreeVector maxpt(bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());
  ThreeVector span = maxpt - basept;
  if(span.x<=0) scale.x = 1.0; // Safety
  else          scale.x = 1.0/span.x;
  if(span.y<=0) scale.y = 1.0;
  else          scale.y = 1.0/span.y;
  if(span.y<=0) scale.z = 1.0;
  else          scale.z = 1.0/span.z;

  string cmdoptreq = GetStringInitValue("script_args","relpt");
  command_options.push_back(Oxs_TclCommandLineOption("minpt",3));
  command_options.push_back(Oxs_TclCommandLineOption("maxpt",3));
  command_options.push_back(Oxs_TclCommandLineOption("span",3));
  command_options.push_back(Oxs_TclCommandLineOption("rawpt",3));
  command_options.push_back(Oxs_TclCommandLineOption("relpt",3));

  cmd.SetBaseCommand(InstanceName(),
		     director->GetMifInterp(),
		     GetStringInitValue("script"),
		     Oxs_ParseTclCommandLineRequest(InstanceName(),
						    command_options,
						    cmdoptreq));
  // Fill fixed command line args
  int index;
  index = command_options[0].position; // minpt
  if(index>=0) {
    cmd.SetCommandArg(index,basept.x);
    cmd.SetCommandArg(index+1,basept.y);
    cmd.SetCommandArg(index+2,basept.z);
  }
  index = command_options[1].position; // maxpt
  if(index>=0) {
    cmd.SetCommandArg(index,maxpt.x);
    cmd.SetCommandArg(index+1,maxpt.y);
    cmd.SetCommandArg(index+2,maxpt.z);
  }
  index = command_options[2].position; // span
  if(index>=0) {
    cmd.SetCommandArg(index,span.x);
    cmd.SetCommandArg(index+1,span.y);
    cmd.SetCommandArg(index+2,span.z);
  }

  VerifyAllInitArgsUsed();
}

BOOL Oxs_ScriptScalarField::Init()
{ return 1; }

Oxs_ScriptScalarField::~Oxs_ScriptScalarField()
{}

REAL8m
Oxs_ScriptScalarField::Value
(const ThreeVector& pt) const
{
  cmd.SaveInterpResult();

  // Fill variable command line args
  int index;
  index = command_options[3].position; // rawpt
  if(index>=0) {
    cmd.SetCommandArg(index,pt.x);
    cmd.SetCommandArg(index+1,pt.y);
    cmd.SetCommandArg(index+2,pt.z);
  }
  index = command_options[4].position; // relpt
  if(index>=0) {
    cmd.SetCommandArg(index,(pt.x-basept.x)*scale.x);
    cmd.SetCommandArg(index+1,(pt.y-basept.y)*scale.y);
    cmd.SetCommandArg(index+2,(pt.z-basept.z)*scale.z);
  }

  cmd.Eval();

  if(cmd.GetResultListSize()!=1) {
    string msg
      = string("Return script value is not a single scalar: ")
      + cmd.GetWholeResult();
    cmd.RestoreInterpResult();
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  double result;
  cmd.GetResultListItem(0,result);

  cmd.RestoreInterpResult();

  return static_cast<REAL8m>(result);
}

void
Oxs_ScriptScalarField::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<REAL8m>& array) const
{
  DefaultFillMeshValue(mesh,array);
}
