/* FILE: timeevolver.cc                 -*-Mode: c++-*-
 *
 * Abstract time evolver class
 *
 */

#include "director.h"
#include "timeevolver.h"

/* End includes */

// Constructors
Oxs_TimeEvolver::Oxs_TimeEvolver
(const char* name,     // Child instance id
 Oxs_Director* newdtr) // App director
  : Oxs_Evolver(name,newdtr), energy_calc_count(0),
    total_energy_output(-5),total_field_output(-5),
    energy_calc_count_output(-5)
{}

Oxs_TimeEvolver::Oxs_TimeEvolver
(const char* name,
 Oxs_Director* newdtr,
 const char* argstr)      // MIF block argument string
  : Oxs_Evolver(name,newdtr,argstr), energy_calc_count(0),
    total_energy_output(-5),total_field_output(-5),
    energy_calc_count_output(-5)
{}


// GetEnergyDensity: Note that mxH is returned, as opposed to MxH.
// This relieves this routine from needing to know what Ms is, and saves
// an unneeded multiplication (since the evolver is just going to divide
// it back out again to calculate dm/dt (as opposed again to dM/dt)).
// The returned energy array is average energy density for the
// corresponding cell in J/m^3; mxH is in A/m, pE_pt (partial derivative
// of E with respect to t) is in J/s.  Any of mxH or H may be
// NULL, which disables assignment for that variable.
void Oxs_TimeEvolver::GetEnergyDensity
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>* mxH_req,
 Oxs_MeshValue<ThreeVector>* H_req,
 REAL8m& pE_pt)
{
  // Update call count
  ++energy_calc_count;

  // Allocate scratch space as needed
  temp_energy.AdjustSize(state.mesh);
  temp_field.AdjustSize(state.mesh);
  energy.AdjustSize(state.mesh);
  if(mxH_req!=NULL) mxH_req->AdjustSize(state.mesh);
  if(H_req!=NULL) H_req->AdjustSize(state.mesh);

  Oxs_MeshValue<ThreeVector>* H_fill = H_req;
  Oxs_MeshValue<ThreeVector>* field_cache = NULL;
  if(total_field_output.GetCacheRequestCount()>0) {
    total_field_output.cache.state_id=0;
    field_cache = &total_field_output.cache.value;
    field_cache->AdjustSize(state.mesh);
    if(H_fill==NULL) H_fill = field_cache;
  }
  /// If field is requested by both H_req and total_field_output,
  /// then fill H_req first, and copy to field_cache at end.

  // For convenience
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;

  UINT4m i;
  UINT4m vecsize = state.mesh->Size();
  UINT4m energy_obj_count = director->GetEnergyObjCount();
  UINT4m energy_obj_index = 0;
  pE_pt=0.;
  if(energy_obj_index<energy_obj_count) {
    Oxs_EnergyData oed(state);
    oed.energy_buffer = &temp_energy;
    oed.field_buffer  = &temp_field;
    director->GetEnergyObj(energy_obj_index)->GetEnergyData(state,oed);
    if(oed.pE_pt.IsSet()) pE_pt = oed.pE_pt;
    energy = *oed.energy;
    if(H_fill!=NULL) *H_fill = *oed.field;
    if(mxH_req!=NULL) {
      Oxs_MeshValue<ThreeVector>& mxH = *mxH_req;
      const Oxs_MeshValue<ThreeVector>& Hfilled = *oed.field;
      for(i=0;i<vecsize;++i) {
	mxH[i] = -1 * Hfilled[i]; // mxH = -Hxm
	mxH[i] ^= spin[i];
      }
    }
    while(++energy_obj_index<energy_obj_count) {
      Oxs_EnergyData oed(state);
      oed.energy_buffer = &temp_energy;
      oed.field_buffer  = &temp_field;
      director->GetEnergyObj(energy_obj_index)->GetEnergyData(state,oed);
      if(oed.pE_pt.IsSet()) pE_pt += oed.pE_pt;
      energy += *oed.energy;
      if(H_fill!=NULL) *H_fill += *oed.field;
      if(mxH_req!=NULL) {
	Oxs_MeshValue<ThreeVector>& mxH = *mxH_req;
	const Oxs_MeshValue<ThreeVector>& Hfilled = *oed.field;
	ThreeVector torque;
	for(i=0;i<vecsize;++i) {
	  torque = spin[i];
	  torque ^= Hfilled[i];
	  mxH[i] += torque;
	}
      }
    }
  } else {
    // No energy terms!
    energy = 0.0;
    if(H_fill !=NULL) *H_fill = ThreeVector(0.,0.,0.);
    if(mxH_req!=NULL) *mxH_req = ThreeVector(0.,0.,0.);
  }

  if(field_cache!=NULL) {
    total_field_output.cache.state_id=state.Id();
    if(field_cache!=H_fill) {
      // Field requested by both H_req and total_field_output,
      // so copy from H_req to field_cache.
      *field_cache = *H_fill;
    }
    field_cache=NULL; // Safety
  }

  // Calculate and store total energy sum if output object
  // total_energy_output has cache enabled.
  if (total_energy_output.GetCacheRequestCount()>0) {
    const Oxs_Mesh* mesh = state.mesh;
    REAL8m energy_sum = 0.0;
    for(i=0;i<vecsize;++i) {
      energy_sum += energy[i]*mesh->Volume(i);
    }
    total_energy_output.cache.state_id=state.Id();
    total_energy_output.cache.value=energy_sum;
  }

}

void Oxs_TimeEvolver::UpdateEnergyOutputs(const Oxs_SimState& state)
{
  if(state.Id()==0) { // Safety
    return;
  }
  Oxs_MeshValue<REAL8m> energy(state.mesh);
  REAL8m pE_pt;
  GetEnergyDensity(state,energy,NULL,NULL,pE_pt);
}

void Oxs_TimeEvolver::FillEnergyCalcCountOutput(const Oxs_SimState& state)
{
  energy_calc_count_output.cache.state_id=state.Id();
  energy_calc_count_output.cache.value
    = static_cast<REAL8m>(energy_calc_count);
}

BOOL Oxs_TimeEvolver::Init()
{
  total_energy_output.Setup(this,InstanceName(),
                           "Total energy","J",1,
                           &Oxs_TimeEvolver::UpdateEnergyOutputs,
                           director);
  total_field_output.Setup(this,InstanceName(),
                           "Total field","A/m",1,
                           &Oxs_TimeEvolver::UpdateEnergyOutputs,
                           director);
  energy_calc_count_output.Setup(this,InstanceName(),
                           "Energy calc count","",0,
                           &Oxs_TimeEvolver::FillEnergyCalcCountOutput,
                           director);
  /// Note: MSVC++ 6.0 requires fully qualified member names

  // Eventually, caching should be handled by controlling Tcl script.
  // Until then, request caching of scalar energy output by default.
  total_energy_output.CacheRequestIncrement(1);

  energy_calc_count=0;
  // Should scratch space be release here?
  return 1;
}
